import __local__
from luma.reduction.manifold import SammonMapping

from sklearn.datasets import make_swiss_roll
import matplotlib.pyplot as plt


X, y = make_swiss_roll(n_samples=500, noise=0.3)

model = SammonMapping(
    n_components=2, max_iter=100, max_halves=35, initialize="pca", verbose=True
)
Z = model.fit_transform(X)

fig = plt.figure(figsize=(10, 5))
ax1 = fig.add_subplot(1, 2, 1, projection="3d")
ax2 = fig.add_subplot(1, 2, 2)

ax1.scatter(X[:, 0], X[:, 1], X[:, 2], c=y, cmap="rainbow")
ax1.set_xlabel(r"$x$")
ax1.set_ylabel(r"$y$")
ax1.set_zlabel(r"$z$")
ax1.set_title("Original S-Curve")

ax2.scatter(Z[:, 0], Z[:, 1], c=y, cmap="rainbow")
ax2.set_xlabel(r"$z_1$")
ax2.set_ylabel(r"$z_2$")
ax2.set_title(f"After {type(model).__name__}")
ax2.grid(alpha=0.2)

plt.tight_layout()
plt.show()



from collections import Counter, defaultdict
from itertools import chain
from typing import Generator, Tuple, Union
import numpy as np

from luma.interface.util import Matrix, Vector

FoldType = Union["KFold", "StratifiedKFold", "GroupKFold"]


__all__ = ("KFold", "StratifiedKFold", "GroupKFold")


class KFold:
    """
    K-Fold cross-validation is a model evaluation method that divides
    the dataset into `k` equal or nearly equal sized folds. In each
    of `k` iterations, one fold is used as the test set, and the
    remaining `k-1` folds are combined to form the training set. This
    process ensures that every data point gets to be in the test set
    exactly once and in the training set `k-1` times. It's widely used
    to assess the performance of a model with limited data, providing
    a robust estimate of its generalization capability.

    Parameters
    ----------
    `X` : Input data
    `y` : Target data
    `n_folds` : Number of folds
    `shuffle` : Whether to shuffle the dataset
    `random_state` : Seed for random shuffling

    Properties
    ----------
    ```py
    @property
    def split(self)
    ```
    This returns a generator with the type:
    ```py
    Generator[Tuple[Vector, Vector], None, None]
    ```
    yielding train indices and test indices.

    Examples
    --------
    Usage of the generator returned by the property `split`:

    ```py
    kfold = KFold(X, y, n_folds=5, shuffle=True)

    for train_indices, test_indices in kfold.split:
        X_train, y_train = X[train_indices], y[train_indices]
        ...
    ```
    """

    def __init__(
        self,
        X: Matrix,
        y: Vector,
        n_folds: int = 5,
        shuffle: bool = True,
        random_state: int = None,
    ) -> None:
        self.X = X
        self.y = y
        self.shuffle = shuffle
        self.random_state = random_state
        self.n_folds = n_folds

        self.set_param_ranges({"n_folds": ("0,+inf", int)})
        self.check_param_ranges()

    @property
    def split(self) -> Generator[Tuple[Vector, Vector], None, None]:
        np.random.seed(self.random_state)
        m, _ = self.X.shape
        fold_size = m // self.n_folds

        indices = np.arange(m)
        if self.shuffle:
            np.random.shuffle(indices)

        for i in range(self.n_folds):
            start = i * fold_size
            end = start + fold_size if i < self.n_folds - 1 else m

            test_indices = indices[start:end]
            train_indices = np.concatenate((indices[:start], indices[end:]))

            yield train_indices, test_indices


class StratifiedKFold:
    """
    Stratified K-Fold cross-validation is a model evaluation method similar
    to KFold but it divides the dataset in a way that preserves the
    percentage of samples for each class. This is especially useful for
    handling imbalances in the dataset. It ensures each fold is a good
    representative of the whole by maintaining the same class distribution
    in each fold as in the complete dataset.

    Parameters
    ----------
    `X` : Input data
    `y` : Target data
    `n_folds` : Number of folds
    `shuffle` : Whether to shuffle the dataset
    `random_state` : Seed for random shuffling

    Properties
    ----------
    ```py
    @property
    def split(self)
    ```
    This returns a generator with the type:
    ```py
    Generator[Tuple[Vector, Vector], None, None]
    ```
    yielding train indices and test indices.

    Examples
    --------
    Usage of the generator returned by the property `split`:

    ```py
    kfold = StratifiedKFold(X, y, n_folds=5, shuffle=True)

    for train_indices, test_indices in kfold.split:
        X_train, y_train = X[train_indices], y[train_indices]
        ...
    ```
    """

    def __init__(
        self,
        X: Matrix,
        y: Vector,
        n_folds: int = 5,
        shuffle: bool = True,
        random_state: int = None,
    ) -> None:
        self.X = X
        self.y = y
        self.n_folds = n_folds
        self.shuffle = shuffle
        self.random_state = random_state

        self.set_param_ranges({"n_folds": ("0,+inf", int)})
        self.check_param_ranges()

    @property
    def split(self) -> Generator[Tuple[Vector, Vector], None, None]:
        np.random.seed(self.random_state)

        indices_per_class = defaultdict(list)
        for idx, class_ in enumerate(self.y):
            indices_per_class[class_].append(idx)

        if self.shuffle:
            for indices in indices_per_class.values():
                np.random.shuffle(indices)

        folds = defaultdict(list)
        for class_, indices in indices_per_class.items():
            fold_size = len(indices) // self.n_folds
            for i in range(self.n_folds):
                start = i * fold_size
                end = start + fold_size if i < self.n_folds - 1 else len(indices)
                folds[i].extend(indices[start:end])

        if self.shuffle:
            for fold_indices in folds.values():
                np.random.shuffle(fold_indices)

        for i in range(self.n_folds):
            test_indices = folds[i]
            train_indices = list(
                chain.from_iterable(folds[j] for j in range(self.n_folds) if j != i)
            )

            yield Vector(train_indices), Vector(test_indices)


class GroupKFold:
    """
    Group K-Fold cross-validation provides train/test indices to split data into
    train/test sets. Each set contains group samples that are entirely in the set
    of training or test. This cross-validation object is a variation of KFold
    that ensures samples from the same group are not split between training and
    testing sets.

    Parameters
    ----------
    `X` : Input data
    `y` : Target data
    `groups` : Array of group labels for the samples
    `n_folds` : Number of folds. Must be at least 2.
    `shuffle` : Whether to shuffle the dataset
    `random_state` : Seed for random shuffling

    Properties
    ----------
    ```py
    @property
    def split(self)
    ```
    This returns a generator with the type:
    ```py
    Generator[Tuple[Vector, Vector], None, None]
    ```
    yielding train indices and test indices.

    Examples
    --------
    Usage of the generator returned by the property `split`:

    ```py
    kfold = GroupKFold(X, y, n_folds=5, shuffle=True)

    for train_indices, test_indices in kfold.split:
        X_train, y_train = X[train_indices], y[train_indices]
        ...
    ```
    """

    def __init__(
        self,
        X: Matrix,
        y: Vector,
        groups: Vector,
        n_folds: int = 5,
        shuffle: bool = True,
        random_state: int = None,
    ) -> None:
        self.X = X
        self.y = y
        self.groups = groups
        self.n_folds = n_folds
        self.shuffle = shuffle
        self.random_state = random_state

        self.set_param_ranges({"n_folds": ("0,+inf", int)})
        self.check_param_ranges()

    @property
    def split(self) -> Generator[Tuple[Vector, Vector], None, None]:
        np.random.seed(self.random_state)
        _, groups_indices = np.unique(self.groups, return_inverse=True)

        group_counts = Counter(groups_indices)
        group_indices = defaultdict(list)
        for idx, group_idx in enumerate(groups_indices):
            group_indices[group_idx].append(idx)

        sorted_groups = sorted(group_counts, key=group_counts.get, reverse=True)

        if self.shuffle:
            np.random.shuffle(sorted_groups)

        folds = [[] for _ in range(self.n_folds)]
        for group in sorted_groups:
            smallest = min(range(self.n_folds), key=lambda x: len(folds[x]))
            folds[smallest].extend(group_indices[group])

        for i in range(self.n_folds):
            test_indices = folds[i]
            train_indices = list(set(range(len(self.X))) - set(test_indices))

            yield Vector(train_indices), Vector(test_indices)



from typing import Any, Dict, List, TypeVar
from scipy.stats import rv_continuous, rv_discrete
import numpy as np
import random

from luma.interface.util import Matrix, Vector
from luma.core.super import Estimator, Evaluator, Optimizer
from luma.interface.exception import NotFittedError
from luma.model_selection.cv import CrossValidator
from luma.model_selection.fold import FoldType, KFold

DT = TypeVar("DT", bound=rv_continuous | rv_discrete)


__all__ = ("GridSearchCV", "RandomizedSearchCV")


class GridSearchCV(Optimizer):
    """
    Grid seach with cross validation(CV) is a method used to tune hyperparameters
    of a model by searching through a specified parameter grid. It evaluates all
    combinations of parameter values to find the best combination. This process
    is performed using cross-validation to ensure model performance is not biased
    to a specific data split. The result is the optimal set of parameters that
    yield the best model performance according to a chosen metric.

    Parameters
    ----------
    `estimator` : An estimator to fit and evaluate
    `param_grid` : Parameter grid for repetitive search
    `cv` : K-fold size for cross validation
    `metric` : Scoring metric for evaluation
    `maximize` : Whether to optimize in a way of maximizing certain metric
    `shuffle` : Whether to shuffle the dataset
    `fold_type` : Fold type (Default `KFold`)
    `refit` : Whether to re-fit the estimator with the best parameters found
    `random_state` : Seed for random sampling for cross validation

    Properties
    ----------
    Get the best(optimized) estimator:
        ```py
        @property
        def best_model(self) -> Estimator
        ```

    Notes
    -----
    * An instance of the estimator must be passed to `estimator`
    * For `metric`, both class or instance are possible
    * Only `Pipeline` is allowed for meta estimator

    Examples
    --------
    >>> param_grid = {'param_1': [...],
                      'param_2': [...],
                      ...,
                      AnyStr: List[Any]}

    >>> grid = GridSearchCV(estimator=AnyEstimator(),
                            param_grid=param_grid,
                            cv=5,
                            metric=AnyEvaluator,
                            refit=True,
                            random_state=None)
    >>> grid.fit(X, y)
    >>> score = grid.best_score
    >>> model = grid.best_model

    """

    def __init__(
        self,
        estimator: Estimator,
        param_grid: dict,
        cv: int = 5,
        metric: Evaluator = None,
        maximize: bool = True,
        refit: bool = True,
        shuffle: bool = True,
        fold_type: FoldType = KFold,
        random_state: int = None,
        verbose: bool = False,
    ) -> None:
        self.estimator = estimator
        self.param_grid = param_grid
        self.cv = cv
        self.metric = metric
        self.maximize = maximize
        self.refit = refit
        self.shuffle = shuffle
        self.fold_type = fold_type
        self.random_state = random_state
        self.verbose = verbose
        self.scores_ = []
        self._fitted = False

        self.set_param_ranges({"cv": ("0,+inf", int)})
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> Estimator:
        best_score = None
        best_params = None
        param_combinations = self._get_param_combinations()
        max_iter = len(param_combinations)

        if self.verbose:
            print(
                f"Fitting {self.cv} folds for {max_iter} candidates,",
                f"totalling {self.cv * max_iter} fits.\n",
            )

        for i, params in enumerate(param_combinations, start=1):
            self._set_params(params)
            cv_model = CrossValidator(
                estimator=self.estimator,
                metric=self.metric,
                cv=self.cv,
                shuffle=self.shuffle,
                fold_type=self.fold_type,
                random_state=self.random_state,
                verbose=self.verbose,
            )

            _, mean_score = cv_model.score(X, y)
            self.scores_.append((params, mean_score))

            if self.verbose:
                params_f = {
                    k: f"{v:.4f}" if isinstance(v, float) else v
                    for k, v in params.items()
                }
                print(
                    f"[GridSearchCV] candidate {i}/{max_iter} {params_f}",
                    f"- score:{mean_score:.3f}",
                )

            if self.maximize:
                if best_score is None or mean_score > best_score:
                    best_score = mean_score
                    best_params = params
            else:
                if best_score is None or mean_score < best_score:
                    best_score = mean_score
                    best_params = params

        self.best_score = best_score
        self.best_params = best_params

        if self.verbose:
            print(f"\n[GridSearchCV] Best params: {self.best_params}")
            print(f"[GridSearchCV] Best score: {self.best_score}")

        if self.refit:
            self.estimator = type(self.estimator)()
            self._set_params(best_params)
            self.estimator.fit(X, y)

        self._fitted = True
        return self.best_model

    def _get_param_combinations(self) -> List[Dict[str, Any]]:
        keys, values = zip(*self.param_grid.items())
        param_combinations = Matrix(np.meshgrid(*values)).T.reshape(-1, len(keys))
        param_combinations = [dict(zip(keys, v)) for v in param_combinations]
        return param_combinations

    def _set_params(self, params: dict) -> None:
        self.estimator.set_params(**params)

    @property
    def best_model(self) -> Estimator:
        if not self._fitted:
            raise NotFittedError(self)
        return self.estimator


class RandomizedSearchCV(Optimizer):
    """
    Randomized search with cross-validation (CV) is a method used to tune
    hyperparameters of a model by searching through a specified parameter
    distribution. It randomly samples combinations of parameter values to find
    the best combination. This process is performed using cross-validation to
    ensure model performance is not biased to a specific data split. The result
    is the optimal set of parameters that yield the best model performance
    according to a chosen metric.

    Parameters
    ----------
    `estimator` : An estimator to fit and evaluate
    `param_dist` : Parameter distributions for random search
    `max_iter` : Number of parameter settings that are sampled
    `cv` : K-fold size for cross-validation
    `metric` : Scoring metric for evaluation
    `maximize` : Whether to optimize in a way of maximizing certain metric
    `refit` : Whether to re-fit the estimator with the best parameters found
    `shuffle` : Whether to shuffle the dataset
    `fold_type` : Fold type (Default `KFold`)
    `random_state` : Seed for random sampling for cross-validation and random search
    `verbose` : Whether to print progress messages

    Properties
    ----------
    Get the best(optimized) estimator:
        ```py
        @property
        def best_model(self) -> Estimator
        ```

    Notes
    -----
    * An instance of the estimator must be passed to `estimator`
    * For `metric`, both class or instance are possible
    * Only `Pipeline` is allowed for meta estimator
    * Type `DT` is a generic type for `scipy`'s distribution types
        (e.g. `rv_continuous`, `rv_discrete`)

    Examples
    --------
    >>> param_dist = {'param_1': [...],
                      'param_2': [...],
                      ...,
                      AnyStr: Vector | DT }

    >>> rand = RandomizedSearchCV(estimator=AnyEstimator(),
                                  param_dist=param_dist,
                                  max_iter=100,
                                  cv=5,
                                  metric=AnyEvaluator,
                                  refit=True,
                                  random_state=None)
    >>> rand.fit(X, y)
    >>> score = rand.best_score
    >>> model = rand.best_model

    """

    def __init__(
        self,
        estimator: Estimator,
        param_dist: Dict[str, Vector | DT],
        max_iter: int = 100,
        cv: int = 5,
        metric: Evaluator = None,
        maximize: bool = True,
        refit: bool = True,
        shuffle: bool = True,
        fold_type: FoldType = KFold,
        random_state: int = None,
        verbose: bool = False,
    ) -> None:
        self.estimator = estimator
        self.param_dist = param_dist
        self.max_iter = max_iter
        self.cv = cv
        self.metric = metric
        self.maximize = maximize
        self.refit = refit
        self.shuffle = shuffle
        self.fold_type = fold_type
        self.random_state = random_state
        self.verbose = verbose
        self.scores_ = []
        self._fitted = False

        self.set_param_ranges({"cv": ("0,+inf", int), "max_iter": ("0<,+inf", int)})
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> Estimator:
        best_score = None
        best_params = None

        if self.verbose:
            print(
                f"Fitting {self.cv} folds for {self.max_iter} candidates,",
                f"toatlling {self.cv * self.max_iter} fits.\n",
            )

        for i in range(self.max_iter):
            params = self._sample_params()
            self._set_params(params)

            cv_model = CrossValidator(
                estimator=self.estimator,
                metric=self.metric,
                cv=self.cv,
                shuffle=self.shuffle,
                fold_type=self.fold_type,
                random_state=self.random_state,
                verbose=self.verbose,
            )

            _, mean_score = cv_model.score(X, y)
            self.scores_.append((params, mean_score))

            if self.verbose:
                params_f = {
                    k: f"{v:.4f}" if isinstance(v, float) else v
                    for k, v in params.items()
                }
                print(
                    f"[RandomSearchCV] candidate {i + 1}/{self.max_iter}",
                    f"{params_f} - score: {mean_score:.3f}",
                )

            if self.maximize:
                if best_score is None or mean_score > best_score:
                    best_score = mean_score
                    best_params = params
            else:
                if best_score is None or mean_score < best_score:
                    best_score = mean_score
                    best_params = params

        self.best_score = best_score
        self.best_params = best_params

        if self.verbose:
            print(f"\n[RandomSearchCV] Best params: {self.best_params}")
            print(f"[RandomSearchCV] Best score: {self.best_score}")

        if self.refit:
            self.estimator = type(self.estimator)()
            self._set_params(best_params)
            self.estimator.fit(X, y)

        self._fitted = True
        return self.best_model

    def _set_params(self, params: dict) -> None:
        self.estimator.set_params(**params)

    def _sample_params(self) -> Dict[str, Any]:
        params = {}
        for key, dist in self.param_dist.items():
            if hasattr(dist, "rvs"):
                params[key] = dist.rvs()
            else:
                params[key] = random.choice(dist)

        return params

    @property
    def best_model(self) -> Estimator:
        if not self._fitted:
            raise NotFittedError(self)
        return self.estimator



from typing import Literal, Optional
from scipy.spatial.distance import pdist, squareform
from scipy.linalg import eigh
import matplotlib.pyplot as plt
import numpy as np

from luma.clustering.kmeans import KMeansClusteringPlus
from luma.clustering.hierarchy import AgglomerativeClustering, DivisiveClustering
from luma.interface.util import Matrix
from luma.core.super import Estimator, Evaluator, Unsupervised
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.metric.clustering import SilhouetteCoefficient


__all__ = (
    "SpectralClustering",
    "NormalizedSpectralClustering",
    "HierarchicalSpectralClustering",
    "AdaptiveSpectralClustering",
)


class SpectralClustering(Estimator, Unsupervised):
    """
    Spectral clustering transforms data into a new space using the
    eigenvectors of its graph Laplacian, revealing cluster structure.
    It constructs a similarity graph from the data, capturing complex
    relationships between points. The algorithm then applies
    dimensionality reduction followed by a traditional clustering
    method, like K-means, in this transformed space. This approach
    is effective for identifying clusters with irregular shapes
    and non-linearly separable data.

    Parameters
    ----------
    `n_clusters` : Number of clusters to estimate
    `gamma` : Scaling factor for Gaussian kernel

    Examples
    --------
    >>> sp = SpectralClustering()
    >>> sp.fit(X, y)
    >>> lables = sp.labels # Get assigned labels

    """

    def __init__(self, n_clusters: int, gamma: float = 1.0):
        self.n_clusters = n_clusters
        self.gamma = gamma
        self._X = None
        self._fitted = False

        self.set_param_ranges(
            {"n_clusters": ("0<,+inf", int), "gamma": ("0<,+inf", None)}
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> "SpectralClustering":
        self._X = X
        W = self._similarity_matrix(X)
        L = self._laplacian(W)

        _, eigenvecs = eigh(L)
        V = eigenvecs[:, : self.n_clusters]
        kmeans = KMeansClusteringPlus(n_clusters=self.n_clusters)
        self.kmeans = kmeans.fit(V)

        self._fitted = True
        return self

    def _similarity_matrix(self, X: Matrix) -> Matrix:
        sq_dists = pdist(X, "sqeuclidean")
        mat_sq_dists = squareform(sq_dists)
        return np.exp(-self.gamma * mat_sq_dists)

    def _laplacian(self, W: Matrix) -> Matrix:
        D = np.diag(np.sum(W, axis=1))
        return D - W

    @property
    def labels(self) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return self.kmeans.labels

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)


class NormalizedSpectralClustering(Estimator, Unsupervised):
    """
    Normalized spectral clustering is a technique that begins by constructing
    a similarity matrix from the data, often using a Gaussian kernel. The graph
    Laplacian matrix, derived from this similarity matrix, is then normalized
    to enhance the distinctiveness of clusters. Eigenvalue decomposition is
    applied to this normalized Laplacian, and the resulting eigenvectors are
    used to transform the data into a new space. Finally, a clustering algorithm,
    such as k-means, is applied in this transformed space, effectively revealing
    clusters that are not easily distinguishable in the original space.

    Parameters
    ----------
    `n_clusters` : Number of clusters to estimate
    `gamma` : Scaling factor for Gaussian kernel
    `strategy` : Normalization strategy (e.g. `symmetric`, `random-walk`)

    """

    def __init__(
        self,
        n_clusters: int,
        gamma: float = 1.0,
        strategy: Literal["symmetric", "random-walk"] = "symmetric",
    ) -> None:
        self.n_clusters = n_clusters
        self.gamma = gamma
        self.strategy = strategy
        self._X = None
        self._fitted = False

        self.set_param_ranges(
            {"n_clusters": ("0<,+inf", int), "gamma": ("0<,+inf", None)}
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> "NormalizedSpectralClustering":
        self._X = X
        W = self._similarity_matrix(X)
        L_norm = self._normalize_laplacian(W)

        _, eigvecs = eigh(L_norm)
        V = eigvecs[:, : self.n_clusters]
        kmeans = KMeansClusteringPlus(n_clusters=self.n_clusters)
        self.kmeans = kmeans.fit(V)

        self._fitted = True
        return self

    def _similarity_matrix(self, X: Matrix) -> Matrix:
        sq_dists = pdist(X, "sqeuclidean")
        mat_sq_dists = squareform(sq_dists)
        return np.exp(-self.gamma * mat_sq_dists)

    def _normalize_laplacian(self, W: Matrix) -> Matrix:
        D = np.diag(np.sum(W, axis=1))
        D_inv = np.diag(1 / np.diag(D))
        D_inv_sqrt = np.diag(1 / np.sqrt(np.diag(D)))

        if self.strategy == "symmetric":
            L_norm = np.eye(*D.shape) - D_inv_sqrt.dot(W).dot(D_inv_sqrt)
        elif self.strategy == "random-walk":
            L_norm = np.eye(*D.shape) - D_inv.dot(W)
        else:
            raise UnsupportedParameterError(self.strategy)

        return L_norm

    @property
    def labels(self) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return self.kmeans.labels

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)


class HierarchicalSpectralClustering(Estimator, Unsupervised):
    """
    Hierarchical spectral clustering starts by applying spectral clustering
    to transform data into a space where clusters are more linearly separable.
    This involves constructing a similarity matrix, computing the Laplacian,
    and performing eigenvalue decomposition. The transformed data is then
    clustered using hierarchical methods, either agglomerative (building
    clusters by merging smaller ones) or divisive (splitting larger clusters
    into smaller ones). This approach reveals multi-level structures in the
    data, uncovering both broad and fine-grained clustering patterns.

    Parameters
    ----------
    `n_clusters` : Number of clusters to estimate
    `method` : Method for hierarchical clustering
    (e.g. `agglomerative`, `divisive`)
    `linkage` : Linkage method for `agglomerative` clustering
    (e.g. `single`, `complete`, `average`, `ward`)
    `gamma` : Scaling factor for Gaussian kernel

    """

    def __init__(
        self,
        n_clusters: int,
        method: Literal["agglomerative", "divisive"],
        linkage: Literal["single", "complete", "average"] = "single",
        gamma: float = 1.0,
    ):
        self.n_clusters = n_clusters
        self.method = method
        self.linkage = linkage
        self.gamma = gamma
        self._hierarchy_model = None
        self._X = None
        self._fitted = False

        self.set_param_ranges(
            {"n_clusters": ("0<,+inf", int), "gamma": ("0<,+inf", None)}
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> "HierarchicalSpectralClustering":
        self._X = X
        W = self._similarity_matrix(X)
        L = self._laplacian(W)

        _, eigenvecs = eigh(L)
        V = eigenvecs[:, : self.n_clusters]

        if self.method == "agglomerative":
            self._hierarchy_model = AgglomerativeClustering()
        elif self.method == "divisive":
            self._hierarchy_model = DivisiveClustering()
        else:
            raise UnsupportedParameterError(self.method)

        self._hierarchy_model.n_clusters = self.n_clusters
        if hasattr(self._hierarchy_model, "linkage"):
            self._hierarchy_model.linkage = self.linkage

        self._hierarchy_model.fit(V)

        self._fitted = True
        return self

    def _similarity_matrix(self, X: Matrix) -> Matrix:
        sq_dists = pdist(X, "sqeuclidean")
        mat_sq_dists = squareform(sq_dists)
        return np.exp(-self.gamma * mat_sq_dists)

    def _laplacian(self, W: Matrix) -> Matrix:
        D = np.diag(np.sum(W, axis=1))
        return D - W

    def plot_dendrogram(
        self, ax: Optional[plt.Axes], hide_indices: bool = True, show: bool = False
    ) -> plt.Axes:
        if not self.method == "agglomerative":
            raise UnsupportedParameterError(self.method)

        return self._hierarchy_model.plot_dendrogram(ax, hide_indices, show)

    @property
    def labels(self) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return self._hierarchy_model.labels

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)


class AdaptiveSpectralClustering(Estimator, Unsupervised):
    """
    Adaptive Spectral Clustering automatically determines the number of clusters
    and scale parameters directly from the data, reducing the need for manual
    tuning. It constructs a similarity matrix, typically using a Gaussian kernel
    with an adaptively chosen scale. Eigenvalue decomposition on the Laplacian of
    this matrix reveals the data's structure, and the eigen-gap heuristic is
    employed to select the optimal number of clusters.

    Parameters
    ----------
    `gamma` : Scaling factor for Gaussian kernel
    `max_clusters` : Upper-bound of the number of clusters to estimate

    """

    def __init__(self, gamma: float = 1.0, max_clusters: int = 10) -> None:
        self.gamma = gamma
        self.max_clusters = max_clusters
        self._X = None
        self._fitted = False

        self.set_param_ranges(
            {"max_clusters": ("0<,+inf", int), "gamma": ("0<,+inf", None)}
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> "AdaptiveSpectralClustering":
        self._X = X
        W = self._similarity_matrix(X)
        L = self._laplacian(W)

        eigvals, eigvecs = eigh(L)
        self.n_clusters = self._optimal_clusters(eigvals)

        V = eigvecs[:, : self.n_clusters]
        kmeans = KMeansClusteringPlus(n_clusters=self.n_clusters)
        self.kmeans = kmeans.fit(V)

        self.W = W
        self.L = L
        self._fitted = True
        return self

    def _similarity_matrix(self, X: Matrix) -> Matrix:
        sq_dists = pdist(X, metric="sqeuclidean")
        mat_sq_dists = squareform(sq_dists)
        return np.exp(-self.gamma * mat_sq_dists)

    def _laplacian(self, W: Matrix) -> Matrix:
        D = np.diag(np.sum(W, axis=0))
        return D - W

    def _optimal_clusters(self, eigvals: Matrix) -> Matrix:
        eigval_diff = np.diff(eigvals)
        return np.argmax(eigval_diff[: self.max_clusters]) + 1

    @property
    def labels(self) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return self.kmeans.labels

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)



from typing import Literal
import numpy as np

from luma.interface.util import Matrix, Vector, Scalar, KernelUtil
from luma.core.super import Estimator, Evaluator, Unsupervised
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.metric.clustering import SilhouetteCoefficient


__all__ = (
    "AffinityPropagation",
    "AdaptiveAffinityPropagation",
    "KernelAffinityPropagation",
)


class AffinityPropagation(Estimator, Unsupervised):
    """
    Affinity Propagation is a clustering algorithm that identifies exemplars
    among data points, effectively determining cluster centers. It works by
    exchanging messages between pairs of data points until a set of exemplars
    and corresponding clusters emerges. The algorithm doesn't require the
    number of clusters to be specified in advance. Affinity Propagation is
    particularly useful for problems where the optimal number of clusters
    is unknown or hard to estimate.

    Parameters
    ----------
    `max_iter` : Maximum iterations for message exchange
    `damping` : Balancing factor for ensuring numerical stability and convergence
    `tol` : Early-stopping threshold
    `preference` : Parameter which determines the selectivity of choosing exemplars

    * Self-Similarity Setting:
        The preference value represents the "self-similarity" of each data point.
        It is the value on the diagonal of the similarity matrix, which influences
        how likely a data point is to be chosen as an exemplar (a cluster center).

    * Influencing Cluster Count:
        A higher preference value suggests a greater likelihood for a data point to
        become an exemplar, potentially leading to more clusters. Conversely, a lower
        preference value tends to produce fewer clusters, as fewer points are strong
        candidates for being exemplars.

    * Default Setting:
        If not explicitly set, the preference is often chosen automatically based on
        the input data. Common practices include setting it to the median or mean of
        the similarity values.

    """

    def __init__(
        self,
        max_iter: int = 100,
        damping: float = 0.7,
        preference: Scalar | Vector | Literal["median", "min"] = "median",
        tol: float = None,
        verbose: bool = False,
    ) -> None:
        self.max_iter = max_iter
        self.damping = damping
        self.preference = preference
        self.tol = tol
        self.verbose = verbose
        self._X = None
        self._fitted = False

        self.set_param_ranges({"max_iter": ("0<,+inf", int), "damping": ("0,1", None)})
        self.check_param_ranges()

    def fit(self, X: Matrix) -> "AffinityPropagation":
        self._X = X
        m, _ = X.shape

        S = self._compute_similarity(X)
        S = self._assign_preference(S)
        S += 1e-12 * np.random.normal(size=(m, m)) * (np.max(S) - np.min(S))

        R = np.zeros_like(S)
        A = np.zeros_like(S)

        for iter in range(self.max_iter):
            E_old = R + A
            R = self._compute_responsibility(S, R, A)
            A = self._compute_availability(R, A)
            E_new = R + A

            if self.tol is not None and np.allclose(E_old, E_new, atol=self.tol):
                if not self.verbose:
                    break
                print(f"[AP] Early convergence at iteration {iter}/{self.max_iter}")
                break

            if self.verbose and iter % 10 == 0 and iter:
                print(
                    f"[AP] Finished iteration {iter}/{self.max_iter}",
                    f"with delta-E: {np.linalg.norm(E_new - E_old)}",
                )

        E = R + A
        self._clusters = np.argmax(E, axis=1)
        self.exemplars = np.unique(self._clusters)
        self.centers = X[self.exemplars]

        replace = dict(zip(self.exemplars, range(len(self.exemplars))))
        new_ = np.arange(0, max(self._clusters) + 1)
        new_[list(replace.keys())] = list(replace.values())

        self._clusters = new_[self._clusters]
        self._fitted = True
        return self

    def _compute_similarity(self, X: Matrix) -> Matrix:
        first_ = np.sum(X**2, axis=-1).reshape(-1, 1)
        second_ = np.sum(X**2, axis=-1)
        third_ = -2 * np.dot(X, X.T)

        return -1 * (first_ + second_ + third_)

    def _assign_preference(self, S: Matrix) -> Matrix:
        indices = np.where(~np.eye(S.shape[0], dtype=bool))

        if self.preference == "median":
            pref = np.median(S[indices])
        elif self.preference == "min":
            pref = np.min(S[indices])

        elif isinstance(self.preference, (np.ndarray, float, int)):
            pref = self.preference
        else:
            raise UnsupportedParameterError(self.preference)

        np.fill_diagonal(S, pref)
        return S

    def _compute_responsibility(self, S: Matrix, R: Matrix, A: Matrix) -> Matrix:
        max_ = A + S
        np.fill_diagonal(max_, -np.inf)

        row_indices = np.arange(max_.shape[0])
        max_indices = np.argmax(max_, axis=1)

        row_max = max_[row_indices, max_indices]
        max_[row_indices, max_indices] = -np.inf

        row_max_ = max_[row_indices, np.argmax(max_, axis=1)]
        max_AS = np.zeros_like(S) + row_max.reshape(-1, 1)
        max_AS[row_indices, max_indices] = row_max_

        return (1 - self.damping) * (S - max_AS) + self.damping * R

    def _compute_availability(self, R: Matrix, A: Matrix) -> Matrix:
        R = R.copy()
        diag = np.diag(R).copy()

        np.fill_diagonal(R, 0)
        R = np.where(R < 0, 0, R)

        sum_vectors = np.sum(R, axis=0)
        A_new = np.minimum(0, diag + sum_vectors - R)
        np.fill_diagonal(A_new, np.sum(R, axis=0))

        return (1 - self.damping) * A_new + self.damping * A

    @property
    def labels(self) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return self._clusters

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)


class AdaptiveAffinityPropagation(Estimator, Unsupervised):
    """
    Adaptive Affinity Propagation (AAP) is a clustering algorithm that
    dynamically adjusts the preference parameter during the clustering
    process. It iteratively refines cluster exemplars based on the exchange
    of responsibility and availability messages between data points. The
    adaptation of preferences allows AAP to be more flexible in identifying
    an appropriate number of clusters, especially in complex or ambiguous
    datasets. This approach enhances the standard Affinity Propagation
    algorithm by offering potentially improved clustering accuracy
    and adaptability.

    Parameters
    ----------
    `max_iter` : Maximum iterations for message exchange
    `damping` : Balancing factor for ensuring numerical stability and convergence
    `tol` : Early-stopping threshold
    `lambda_param` : Preference updating factor
    `preference` : Parameter which determines the selectivity of choosing exemplars
    (more details in `AffinityPropagation`)

    """

    def __init__(
        self,
        max_iter: int = 100,
        damping: float = 0.7,
        preference: Scalar | Vector | Literal["median", "min"] = "median",
        lambda_param: float = 0.5,
        tol: float = None,
        verbose: bool = False,
    ) -> None:
        self.max_iter = max_iter
        self.damping = damping
        self.preference = preference
        self.tol = tol
        self.lambda_param = lambda_param
        self.verbose = verbose
        self._X = None
        self._fitted = False

        self.set_param_ranges(
            {
                "max_iter": ("0<,+inf", int),
                "damping": ("0,1", None),
                "lambda_param": ("0,+inf", None),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> "AdaptiveAffinityPropagation":
        self._X = X
        m, _ = X.shape

        S = self._compute_similarity(X)
        S = self._assign_preference(S)
        S += 1e-12 * np.random.normal(size=(m, m)) * (np.max(S) - np.min(S))

        R = np.zeros_like(S)
        A = np.zeros_like(S)

        for iter in range(self.max_iter):
            E_old = R + A
            R = self._compute_responsibility(S, R, A)
            A = self._compute_availability(R, A)
            E_new = R + A

            diag_R = np.diag(R)
            diag_A = np.diag(A)
            self._update_preferences(S, diag_R + diag_A)

            if self.tol is not None and np.allclose(E_old, E_new, atol=self.tol):
                if not self.verbose:
                    break
                print(f"[AAP] Early convergence at iteration {iter}/{self.max_iter}")
                break

            if self.verbose and iter % 10 == 0 and iter:
                print(
                    f"[AAP] Finished iteration {iter}/{self.max_iter}",
                    f"with delta-E: {np.linalg.norm(E_new - E_old)}",
                )

        E = R + A
        self._clusters = np.argmax(E, axis=1)
        self.exemplars = np.unique(self._clusters)
        self.centers = X[self.exemplars]

        replace = dict(zip(self.exemplars, range(len(self.exemplars))))
        new_ = np.arange(0, max(self._clusters) + 1)
        new_[list(replace.keys())] = list(replace.values())

        self._clusters = new_[self._clusters]
        self._fitted = True
        return self

    def _compute_similarity(self, X: Matrix) -> Matrix:
        first_ = np.sum(X**2, axis=-1).reshape(-1, 1)
        second_ = np.sum(X**2, axis=-1)
        third_ = -2 * np.dot(X, X.T)

        return -1 * (first_ + second_ + third_)

    def _assign_preference(self, S: Matrix) -> Matrix:
        indices = np.where(~np.eye(S.shape[0], dtype=bool))

        if self.preference == "median":
            pref = np.median(S[indices])
        elif self.preference == "min":
            pref = np.min(S[indices])

        elif isinstance(self.preference, (np.ndarray, float, int)):
            pref = self.preference
        else:
            raise UnsupportedParameterError(self.preference)

        np.fill_diagonal(S, pref)
        return S

    def _compute_responsibility(self, S: Matrix, R: Matrix, A: Matrix) -> Matrix:
        max_ = A + S
        np.fill_diagonal(max_, -np.inf)

        row_indices = np.arange(max_.shape[0])
        max_indices = np.argmax(max_, axis=1)

        row_max = max_[row_indices, max_indices]
        max_[row_indices, max_indices] = -np.inf

        row_max_ = max_[row_indices, np.argmax(max_, axis=1)]
        max_AS = np.zeros_like(S) + row_max.reshape(-1, 1)
        max_AS[row_indices, max_indices] = row_max_

        return (1 - self.damping) * (S - max_AS) + self.damping * R

    def _compute_availability(self, R: Matrix, A: Matrix) -> Matrix:
        R = R.copy()
        diag = np.diag(R).copy()

        np.fill_diagonal(R, 0)
        R = np.where(R < 0, 0, R)

        sum_vectors = np.sum(R, axis=0)
        A_new = np.minimum(0, diag + sum_vectors - R)
        np.fill_diagonal(A_new, np.sum(R, axis=0))

        return (1 - self.damping) * A_new + self.damping * A

    def _update_preferences(self, S: Matrix, delta: Vector) -> None:
        if not isinstance(self.preference, (int, float)):
            return
        self.preference += self.lambda_param * delta
        np.fill_diagonal(S, self.preference)

    @property
    def labels(self) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return self._clusters

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)


class KernelAffinityPropagation(Estimator, Unsupervised):
    """
    Kernel-based Affinity Propagation (KAP) employs kernel functions
    to transform data into a higher-dimensional space, enabling effective
    clustering of complex, non-linear datasets. This approach enhances
    the standard Affinity Propagation by uncovering intricate cluster
    structures in transformed feature spaces, suitable for non-linearly
    separable data.

    Parameters
    ----------
    `max_iter` : Maximum iterations for message exchange
    `damping` : Balancing factor for ensuring numerical stability and convergence
    `preference` : Parameter which determines the selectivity of choosing exemplars
    `tol` : Early-stopping threshold
    `kernel`: Kernel method (`rbf`, `sigmoid`, `laplacian` are supported)
    `gamma` : Scaling factor for kernel function

    """

    def __init__(
        self,
        max_iter: int = 100,
        damping: float = 0.7,
        preference: Scalar | Vector | Literal["median", "min"] = "median",
        tol: float = None,
        kernel: KernelUtil.FuncType = "rbf",
        gamma: float = 1.0,
        verbose: bool = False,
    ) -> None:
        self.max_iter = max_iter
        self.damping = damping
        self.preference = preference
        self.tol = tol
        self.kernel = kernel
        self.gamma = gamma
        self.verbose = verbose
        self._X = None
        self._fitted = False

        self.set_param_ranges(
            {
                "max_iter": ("0<,+inf", int),
                "damping": ("0,1", None),
                "gamma": ("0<,+inf", None),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> "KernelAffinityPropagation":
        self._X = X
        m, _ = X.shape

        kernel_func = KernelUtil(self.kernel, gamma=self.gamma).kernel_func
        X_trans = kernel_func(X)

        S = self._compute_similarity(X_trans)
        S = self._assign_preference(S)
        S += 1e-12 * np.random.normal(size=(m, m)) * (np.max(S) - np.min(S))

        R = np.zeros_like(S)
        A = np.zeros_like(S)

        for iter in range(self.max_iter):
            E_old = R + A
            R = self._compute_responsibility(S, R, A)
            A = self._compute_availability(R, A)
            E_new = R + A

            if self.tol is not None and np.allclose(E_old, E_new, atol=self.tol):
                if not self.verbose:
                    break
                print(f"[KAP] Early convergence at iteration {iter}/{self.max_iter}")
                break

            if self.verbose and iter % 10 == 0 and iter:
                print(
                    f"[KAP] Finished iteration {iter}/{self.max_iter}",
                    f"with delta-E: {np.linalg.norm(E_new - E_old)}",
                )

        E = R + A
        self._clusters = np.argmax(E, axis=1)
        self.exemplars = np.unique(self._clusters)
        self.centers = X[self.exemplars]

        replace = dict(zip(self.exemplars, range(len(self.exemplars))))
        new_ = np.arange(0, max(self._clusters) + 1)
        new_[list(replace.keys())] = list(replace.values())

        self._clusters = new_[self._clusters]
        self._fitted = True
        return self

    def _compute_similarity(self, X: Matrix) -> Matrix:
        first_ = np.sum(X**2, axis=-1).reshape(-1, 1)
        second_ = np.sum(X**2, axis=-1)
        third_ = -2 * np.dot(X, X.T)

        return -1 * (first_ + second_ + third_)

    def _assign_preference(self, S: Matrix) -> Matrix:
        indices = np.where(~np.eye(S.shape[0], dtype=bool))

        if self.preference == "median":
            pref = np.median(S[indices])
        elif self.preference == "min":
            pref = np.min(S[indices])

        elif isinstance(self.preference, (np.ndarray, float, int)):
            pref = self.preference
        else:
            raise UnsupportedParameterError(self.preference)

        np.fill_diagonal(S, pref)
        return S

    def _compute_responsibility(self, S: Matrix, R: Matrix, A: Matrix) -> Matrix:
        max_ = A + S
        np.fill_diagonal(max_, -np.inf)

        row_indices = np.arange(max_.shape[0])
        max_indices = np.argmax(max_, axis=1)

        row_max = max_[row_indices, max_indices]
        max_[row_indices, max_indices] = -np.inf

        row_max_ = max_[row_indices, np.argmax(max_, axis=1)]
        max_AS = np.zeros_like(S) + row_max.reshape(-1, 1)
        max_AS[row_indices, max_indices] = row_max_

        return (1 - self.damping) * (S - max_AS) + self.damping * R

    def _compute_availability(self, R: Matrix, A: Matrix) -> Matrix:
        R = R.copy()
        diag = np.diag(R).copy()

        np.fill_diagonal(R, 0)
        R = np.where(R < 0, 0, R)

        sum_vectors = np.sum(R, axis=0)
        A_new = np.minimum(0, diag + sum_vectors - R)
        np.fill_diagonal(A_new, np.sum(R, axis=0))

        return (1 - self.damping) * A_new + self.damping * A

    @property
    def labels(self) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return self._clusters

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)



import numpy as np

from luma.interface.util import Matrix
from luma.core.super import Evaluator


__all__ = ("Accuracy", "Precision", "Recall", "F1Score", "Specificity")


class Accuracy(Evaluator):
    @staticmethod
    def score(y_true: Matrix, y_pred: Matrix) -> float:
        return np.mean(y_true == y_pred)


class Precision(Evaluator):
    @staticmethod
    def score(y_true: Matrix, y_pred: Matrix) -> float:
        true_positives = np.sum((y_true == 1) & (y_pred == 1))
        false_positives = np.sum((y_true == 0) & (y_pred == 1))
        return true_positives / (true_positives + false_positives)


class Recall(Evaluator):
    @staticmethod
    def score(y_true: Matrix, y_pred: Matrix) -> float:
        true_positives = np.sum((y_true == 1) & (y_pred == 1))
        false_negatives = np.sum((y_true == 1) & (y_pred == 0))
        return true_positives / (true_positives + false_negatives)


class F1Score(Evaluator):
    @staticmethod
    def score(y_true: Matrix, y_pred: Matrix) -> float:
        precision = Precision.score(y_true, y_pred)
        recall = Recall.score(y_true, y_pred)
        return 2 * (precision * recall) / (precision + recall)


class Specificity(Evaluator):
    @staticmethod
    def score(y_true: Matrix, y_pred: Matrix) -> float:
        true_negatives = np.sum((y_true == 0) & (y_pred == 0))
        false_positives = np.sum((y_true == 0) & (y_pred == 1))
        return true_negatives / (true_negatives + false_positives)



from typing import Optional
from scipy.spatial.distance import pdist, squareform
import matplotlib.pyplot as plt
import numpy as np

from luma.core.super import Evaluator, Visualizer
from luma.interface.util import Matrix, SilhouetteUtil, DBUtil


__all__ = ("SilhouetteCoefficient", "DaviesBouldin", "Inertia")


class SilhouetteCoefficient(Evaluator, Visualizer):
    """
    The Silhouette Coefficient is a measure used to evaluate the quality of
    clusters in a dataset. It calculates how similar each data point is to its
    own cluster compared to other clusters. The coefficient ranges from -1
    (poorly clustered) to +1 (well clustered), with values near 0 indicating
    overlapping clusters. High average scores across a dataset suggest clear,
    well-separated clusters.

    Parameters
    ----------
    `data` : Original data
    `labels` : Labels assigned by clustering estimator

    Examples
    --------
    With Instantiation
    ```py
        sil = SilhouetteCoefficient(data, labels)
        score = sil.score(data, labels) # compute() is a static method
        sil.plot(...)
    ```

    Without Instantiation
    ```py
        score = SilhouetteCoefficient.score(data, labels)
        SilhouetteCoefficient.plot(...) # Error; plot() is an instance method
    ```

    """

    def __init__(self, data: Matrix, labels: Matrix) -> None:
        self.data = data
        self.labels = labels
        self.dist = squareform(pdist(self.data))

    @staticmethod
    def score(data: Matrix, labels: Matrix) -> Matrix[float]:
        scores = []
        dist = squareform(pdist(data))
        for idx, label in enumerate(labels):
            util = SilhouetteUtil(idx, label, labels, dist)
            a = util.avg_dist_within
            b = util.avg_dist_others
            score = (b - a) / max(a, b) if max(a, b) != 0 else 0
            scores.append(score)

        return np.mean(scores)

    def _individual_silhouette(self) -> Matrix:
        silhouette_values = []
        for idx, label in enumerate(self.labels):
            util = SilhouetteUtil(idx, label, self.labels, self.dist)
            a = util.avg_dist_within
            b = util.avg_dist_others
            score = (b - a) / max(a, b) if max(a, b) != 0 else 0
            silhouette_values.append(score)

        return Matrix(silhouette_values)

    def plot(self, ax: Optional[plt.Axes] = None, show: bool = False) -> plt.Axes:
        if ax is None:
            _, ax = plt.subplots()
            show = True

        sample_silhouette = self._individual_silhouette()
        y_lower = 10

        for i in range(len(np.unique(self.labels))):
            values = sample_silhouette[self.labels == i]
            values.sort()

            size_cluster_i = values.shape[0]
            y_upper = y_lower + size_cluster_i

            color = plt.cm.nipy_spectral(float(i) / len(set(self.labels)))
            ax.fill_betweenx(
                np.arange(y_lower, y_upper),
                0,
                values,
                facecolor=color,
                edgecolor=color,
                alpha=0.7,
            )
            y_lower = y_upper + 10

        ax.set_title("Silhouette Coefficient Plot")
        ax.set_xlabel("Silhouette coefficient values")
        ax.set_ylabel("Cluster label")
        ax.axvline(x=self.score(self.data, self.labels), color="red", linestyle="--")

        ax.set_yticks([])
        ax.set_xticks(np.arange(0.0, 1.2, 0.1))
        plt.tight_layout()

        if show:
            plt.show()
        return ax


class DaviesBouldin(Evaluator):
    """
    The Davies-Bouldin Index (DBI) is a metric for evaluating clustering
    algorithms. It compares the average distance within clusters to the
    distance between clusters. Lower DBI values indicate better clustering,
    with compact and well-separated clusters.

    Parameters
    ----------
    `data` : Original data
    `labels` : Labels assigned by clustering estimator

    """

    @staticmethod
    def score(data: Matrix, labels: Matrix) -> float:
        util = DBUtil(data=data, labels=labels)
        centroids = util.cluster_centroids
        scatter = util.within_cluster_scatter
        separation = util.separation

        n_clusters = len(centroids)
        db_values = np.zeros(n_clusters)

        for i in range(n_clusters):
            max_ratio = 0
            for j in range(n_clusters):
                if i == j:
                    continue
                ratio = (scatter[i] + scatter[j]) / separation[i, j]
                if ratio > max_ratio:
                    max_ratio = ratio

            db_values[i] = max_ratio
        db_index = np.mean(db_values)

        return db_index


class Inertia(Evaluator):
    """
    Inertia in clustering quantifies the compactness of clusters by measuring
    the sum of squared distances between data points and their respective
    cluster centroids. It serves as a key metric to evaluate the quality of
    cluster assignments in algorithms like K-means. A lower inertia value
    indicates more cohesive clusters, where data points are closer to their
    centroids. However, minimizing inertia excessively can lead to overfitting,
    with a large number of clusters.

    Parameters
    ----------
    `data` : Original data
    `centroids`: Centroids(or medoids for certain algorithm)

    """

    @staticmethod
    def score(data: Matrix, centroids: Matrix) -> float:
        sq_dist = (data[:, np.newaxis, :] - centroids[np.newaxis, :, :]) ** 2
        dist = np.sqrt(sq_dist.sum(axis=2))

        closest = np.argmin(dist, axis=1)
        inertia = sum((dist[i, closest[i]] ** 2) for i in range(len(data)))

        return inertia



from typing import Optional
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

from luma.core.super import Visualizer


__all__ = ("CorrelationHeatMap", "CorrelationBar", "JointPlot", "MissingProportion")


class CorrelationHeatMap(Visualizer):
    def __init__(self, data: pd.DataFrame) -> None:
        self.data = data
        self.corr = data.corr()

    def plot(
        self,
        ax: Optional[plt.Axes] = None,
        colorMap: str = "rocket",
        annotate: bool = True,
        colorBar: bool = True,
        show: bool = False,
    ) -> plt.Axes:
        if ax is None:
            n_features = self.data.shape[1]
            size = n_features / 2 if n_features < 20 else 10
            _, ax = plt.subplots(figsize=(size + 1, size))

        sns.heatmap(
            self.corr, ax=ax, cmap=colorMap, annot=annotate, cbar=colorBar, fmt="0.2f"
        )

        ax.set_title("Correlation Heat Map")
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax


class CorrelationBar(Visualizer):
    def __init__(self, data: pd.DataFrame, target: str) -> None:
        self.data = data
        self.target = target

    def plot(self, ax: Optional[plt.Axes] = None, show: bool = False) -> plt.Axes:
        if ax is None:
            _, ax = plt.subplots()
        corr_bar = [
            abs(self.data[col].corr(self.data[self.target])) for col in self.data
        ]
        sns.barplot(x=self.data.columns, y=corr_bar, hue=self.data.columns, ax=ax)

        ax.set_xticklabels(ax.get_xticklabels(), rotation=45)
        ax.set_title(f"Correlations with {self.target}")
        ax.set_xlabel("Features")
        ax.set_ylabel("Correlation")
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax


class JointPlot(Visualizer):
    def __init__(self, data: pd.DataFrame, x: str, y: str) -> None:
        self.data = data
        self.x = x
        self.y = y

    def plot(self, color: str = "tab:blue") -> None:
        sns.jointplot(data=self.data, x=self.x, y=self.y, kind="reg", color=color)
        plt.title(f"{self.x} vs. {self.y}")
        plt.tight_layout()
        plt.show()


class MissingProportion(Visualizer):
    def __init__(self, data: pd.DataFrame):
        self.data = data

    def nan_proportions(self) -> pd.DataFrame:
        nan_props = self.data.isna().mean()
        return nan_props

    def plot(self, ax: Optional[plt.Axes] = None, show: bool = False) -> plt.Axes:
        if ax is None:
            _, ax = plt.subplots()
        nan_props = self.nan_proportions()
        sns.barplot(x=nan_props.index, y=nan_props.values, hue=self.data.columns, ax=ax)

        ax.set_xticklabels(ax.get_xticklabels(), rotation=45)
        ax.set_xlabel("Columns")
        ax.set_ylabel("Proportion")
        ax.set_title("Missing Value Proportions")
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax



from typing import Literal, Optional, Tuple, List
from matplotlib import pyplot as plt
from scipy.spatial.distance import cdist
import numpy as np

from luma.core.super import Estimator, Evaluator, Unsupervised
from luma.interface.util import Matrix, Vector, Scalar
from luma.interface.exception import NotFittedError, NotConvergedError
from luma.interface.exception import UnsupportedParameterError

from luma.metric.distance import Euclidean, Minkowski
from luma.metric.clustering import SilhouetteCoefficient


__all__ = ("DBSCAN", "OPTICS", "DENCLUE", "MeanShiftClustering")


class DBSCAN(Estimator, Unsupervised):
    """
    DBSCAN, short for Density-Based Spatial Clustering of Applications
    with Noise, is a clustering algorithm that groups points in a dataset
    by their proximity and density. It identifies clusters as areas of high
    point density, separated by regions of low density. Points in sparse
    regions are classified as noise. DBSCAN is particularly effective at
    discovering clusters of arbitrary shapes and dealing with outliers.

    Parameters
    ----------
    `epsilon` : Radius of a neighborhood hypersphere
    `min_points` : Minimum required points to form a cluster

    """

    def __init__(
        self,
        epsilon: float = 0.1,
        min_points: int = 5,
        metric: Literal["euclidean", "minkowski"] = "euclidean",
    ) -> None:
        self.epsilon = epsilon
        self.min_points = min_points
        self.metric = metric
        self._X = None
        self._fitted = False

        self.metric_func = None
        if self.metric == "euclidean":
            self.metric_func = Euclidean
        elif self.metric == "minkowski":
            self.metric_func = Minkowski
        else:
            raise UnsupportedParameterError(self.metric)

        self.set_param_ranges(
            {"epsilon": ("0<,+inf", None), "min_points": ("0<,+inf", int)}
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> "DBSCAN":
        self._X = X
        clusters = [0] * X.shape[0]
        curPt = 0

        for i in range(X.shape[0]):
            if clusters[i]:
                continue
            neighbors = self._generate_neighbors(X, idx=i)
            if len(neighbors) < self.min_points:
                clusters[i] = -1
            else:
                curPt += 1
                self._expand_cluster(X, neighbors, clusters, i, curPt)

        self._cluster_labels = clusters
        self._fitted = True
        return self

    def _generate_neighbors(self, X: Matrix, idx: int) -> Matrix:
        neighbors = []
        for i in range(X.shape[0]):
            if self.metric_func.compute(X[idx], X[i]) < self.epsilon:
                neighbors.append(i)

        return Matrix(neighbors)

    def _expand_cluster(
        self, X: Matrix, neighbors: Matrix, clusters: Matrix, idx: int, current: int
    ) -> None:
        i = 0
        clusters[idx] = current

        while i < len(neighbors):
            next = neighbors[i]
            if clusters[next] == -1:
                clusters[next] = current

            elif clusters[next] == 0:
                clusters[next] = current
                next_neighbors = self._generate_neighbors(X, idx=next)

                if len(next_neighbors) > self.min_points:
                    neighbors = np.concatenate([neighbors, next_neighbors], axis=0)

            i += 1

    @property
    def labels(self) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return Matrix(self._cluster_labels)

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)


class OPTICS(Estimator, Unsupervised):
    """
    OPTICS (Ordering Points To Identify the Clustering Structure) is an
    unsupervised learning algorithm for identifying cluster structures
    in spatial  It creates an ordered list of points based on
    core-distance and reachability-distance, allowing it to find
    clusters of varying density. Unlike algorithms like DBSCAN, it
    doesn't require a global density threshold, making it versatile
    for complex datasets. The result is often visualized as a
    reachability plot, revealing the data's clustering hierarchy and
    density variations.

    Parameters
    ----------
    `epsilon` : Radius of neighborhood hypersphere
    `min_points` : Minimum nuber of points to form a cluster
    `threshold` : Threshold for filtering samples with large reachabilities

    """

    def __init__(
        self,
        epsilon: float = 1.0,
        min_points: int = 5,
        threshold: float = 1.5,
        verbose: bool = False,
    ) -> None:
        self.epsilon = epsilon
        self.min_points = min_points
        self.threshold = threshold
        self.verbose = verbose
        self._X = None
        self._fitted = False

        self.set_param_ranges(
            {
                "epsilon": ("0<,+inf", None),
                "min_points": ("0<,+inf", int),
                "threshold": ("0,+inf", None),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> "OPTICS":
        self._X = X
        m, _ = X.shape

        self.processed = np.full(m, False, dtype=bool)
        self.reachability = np.full(m, np.inf)
        self.ordered_points = []

        for i in range(m):
            if self.verbose and i % 50 == 0 and i:
                print(
                    f"[OPTICS] Finished for point {i}/{m}",
                    f"with reachability {self.reachability[i]}",
                )

            if self.processed[i]:
                continue
            seeds = []
            point_neighbors = self._neighbors(X, i)
            core_dist = self._core_distance(point_neighbors)

            self.processed[i] = True
            self.ordered_points.append(i)

            if not np.isinf(core_dist):
                self._update(X, core_dist, i, seeds)
                seeds.sort(key=lambda x: x[1])

                while seeds:
                    next_, _ = seeds.pop(0)
                    self.processed[next_] = True
                    self.ordered_points.append(next_)
                    next_neighbors = self._neighbors(X, next_)
                    core_dist = self._core_distance(next_neighbors)

                    if not np.isinf(core_dist):
                        self._update(X, core_dist, next_, seeds)

        self._fitted = True
        return self

    def _core_distance(self, neighbors: Matrix) -> Matrix | Scalar:
        if len(neighbors) >= self.min_points:
            return sorted(neighbors)[self.min_points - 1]
        return np.inf

    def _neighbors(self, X: Matrix, idx: int) -> Matrix:
        distances = cdist([X[idx]], X)[0]
        return distances[distances <= self.epsilon]

    def _update(self, X: Matrix, core_dist: Scalar, idx: int, seeds: list) -> None:
        distances = cdist([X[idx]], X)[0]
        for i, dist in enumerate(distances):
            if dist <= self.epsilon and not self.processed[i]:
                new_reach_dist = max(core_dist, dist)

                if np.isinf(self.reachability[i]):
                    self.reachability[i] = new_reach_dist
                    seeds.append((i, new_reach_dist))
                elif new_reach_dist < self.reachability[i]:
                    self.reachability[i] = new_reach_dist

    def plot_reachability(
        self,
        ax: Optional[plt.Axes] = None,
        color: str = "royalblue",
        show: bool = False,
    ) -> plt.Axes:
        m = range(len(self.ordered_points))
        vals = self.reachability[self.ordered_points]

        if ax is None:
            _, ax = plt.subplots()
            show = True

        ax.plot(m, vals, color=color)
        ax.fill_between(m, vals, color=color, alpha=0.5)

        ax.set_title("Reachability Plot")
        ax.set_xlim(m[0], m[-1])

        ax.set_xlabel("Order of Points")
        ax.set_ylabel("Reachability Distance")
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax

    @property
    def labels(self) -> Vector:
        if not self._fitted:
            raise NotFittedError(self)
        cluster_labels = np.full_like(self.reachability, -1, dtype=int)
        current_ = 0

        for point in self.ordered_points:
            if self.reachability[point] <= self.threshold:
                cluster_labels[point] = current_
            else:
                current_ += 1

        unique_labels = np.unique(cluster_labels[cluster_labels != -1])
        mapping = {original: new for new, original in enumerate(unique_labels)}
        labels = [mapping[label] if label != -1 else -1 for label in cluster_labels]

        return Vector(labels)

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)


class DENCLUE(Estimator, Unsupervised):
    """
    DENCLUE (Density-Based Clustering) is a machine learning algorithm
    for identifying clusters in large datasets. It uses mathematical
    functions, typically Gaussian kernels, to estimate data density.
    Cluster centers are determined by the peaks of these density functions.
    The algorithm excels in handling clusters of arbitrary shapes and noise,
    but is sensitive to parameter settings.

    Parameters
    ----------
    `h` : Smoothing parameter of local density estimation
    `tol` : Threshold for early convergence
    `max_climb` : Maximum number of climbing process for finding local maxima
    `min_density` : Minimum local densities to be considered
    `sample_weight` : Custom individual weights for sample data

    Reference
    ---------
    Hinneburg, A., & Gabriel, H. H. (2007, September). Denclue 2.0: Fast
    clustering based on kernel density estimation. In International symposium
    on intelligent data analysis (pp. 70-80). Berlin, Heidelberg: Springer
    Berlin Heidelberg.

    """

    def __init__(
        self,
        h: float | Literal["auto"] = "auto",
        tol: float = 1e-3,
        max_climb: int = 100,
        min_density: float = 0.0,
        sample_weight: Vector = None,
    ) -> None:
        self.h = h
        self.tol = tol
        self.max_climb = max_climb
        self.min_density = min_density
        self.sample_weight = sample_weight
        self._X = None
        self._fitted = False

        self.set_param_ranges(
            {"max_climb": ("0<,+inf", int), "min_density": ("0,+inf", None)}
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> "DENCLUE":
        self._m, self._n = X.shape

        attractors: Vector = np.zeros((self._m, self._n))
        rad: Vector = np.zeros((self._m, 1))
        density: Vector = np.zeros((self._m, 1))

        if self.h == "auto":
            self.h = np.std(X) / 5
        if self.sample_weight is None:
            self.sample_weight = np.ones((self._m, 1))

        labels = -np.ones(X.shape[0])
        for i in range(self._m):
            attractors[i], density[i], rad[i] = self._climb_hill(X[i], X)

        adj_list = [[] for _ in range(self._m)]
        for i in range(self._m):
            for j in range(i + 1, self._m):
                diff = np.linalg.norm(attractors[i] - attractors[j])
                if diff <= rad[i] + rad[j]:
                    adj_list[i].append(j)
                    adj_list[j].append(i)

        num_clusters = 0
        for cl in self._find_connected_components(adj_list):
            max_instance = max(cl, key=lambda x: density[x])
            max_density = density[max_instance]

            if max_density >= self.min_density:
                labels[cl] = num_clusters
            num_clusters += 1

        self.labels = labels
        self._fitted = True
        return self

    def _climb_hill(self, x: Vector, X: Matrix) -> Tuple[Vector, float, float]:
        error, prob = 99.0, 0.0
        x1 = np.copy(x)

        iters = 0
        r_new, r_old, r_2old = 0.0, 0.0, 0.0
        while True:
            r_3old, r_2old, r_old = r_2old, r_old, r_new
            x0 = np.copy(x1)
            x1, density = self._step(x0, X)

            error, prob = density - prob, density
            r_new = np.linalg.norm(x1 - x0)
            radius = r_3old + r_2old + r_old + r_new

            iters += 1
            if iters > 3 and error < self.tol:
                break
            if iters == self.max_climb:
                raise NotConvergedError(self)

        return x1, prob, radius

    def _step(self, x: Vector, X: Matrix) -> Tuple[Vector, float]:
        sup_weight = 0.0
        x1 = np.zeros((1, self._n))

        for j in range(self._m):
            kernel = self._kernel_func(x, X[j], self._n)
            kernel = kernel * self.sample_weight[j] / (self.h**self._n)

            sup_weight = sup_weight + kernel
            x1 = x1 + (kernel * X[j])

        x1 = x1 / sup_weight
        density = sup_weight / np.sum(self.sample_weight)

        return x1, density

    def _kernel_func(self, xi: Vector, xj: Vector, deg: int):
        kernel = np.exp(-((np.linalg.norm(xi - xj) / self.h) ** 2) / 2)
        kernel /= (2 * np.pi) ** (deg / 2)

        return kernel

    def _find_connected_components(self, adj_list: List[list]) -> list:
        def __dfs(node: int) -> Vector:
            stack, path = [node], []
            while stack:
                vertex = stack.pop()
                if not visited[vertex]:
                    visited[vertex] = True
                    path.append(vertex)
                    stack.extend(set(adj_list[vertex]) - set(path))

            return Matrix(path)

        clusters = []
        visited = [False] * self._m
        for node in range(self._m):
            if visited[node]:
                continue
            cluster = __dfs(node)
            clusters.append(cluster)

        return clusters

    @property
    def labels(self):
        return self._labels

    @labels.setter
    def labels(self, value):
        self._labels = value

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)


class MeanShiftClustering(Estimator, Unsupervised):
    """
    Mean Shift is a non-parametric, iterative clustering algorithm
    that identifies clusters in a dataset by updating candidate cluster
    centers to the mean of points within a given region (bandwidth).
    The process repeats until convergence, where centers no longer
    significantly shift. It automatically determines the number of
    clusters based on the data distribution. Mean Shift is especially
    effective in discovering clusters of arbitrary shapes and sizes.

    Parameters
    ----------
    `bandwidth` : Window size for kernel density estimation
    `max_iter` : Maximum iteration
    `tol` : Tolerence threshold for early convergence

    """

    def __init__(
        self,
        bandwidth: float = 2.0,
        max_iter: int = 300,
        tol: float = 1e-3,
        verbose: bool = False,
    ) -> None:
        self.bandwidth = bandwidth
        self.max_iter = max_iter
        self.tol = tol
        self.verbose = verbose
        self._X = None
        self._fitted = False

        self.set_param_ranges(
            {"bandwidth": ("0<,+inf", None), "max_iter": ("0<,+inf", int)}
        )
        self.check_param_ranges()

    def fit(self, X: Matrix) -> "MeanShiftClustering":
        self._X = X
        m, _ = X.shape

        centers = np.copy(X)
        for iter in range(self.max_iter):
            new_centers = []
            for i in range(m):
                points = X[np.linalg.norm(X - centers[i], axis=1) < self.bandwidth]
                new_center = points.mean(axis=0)
                new_centers.append(new_center)
            new_centers = Matrix(new_centers)

            diff = np.linalg.norm(new_centers - centers)
            if self.verbose and iter % 10 == 0 and iter:
                print(f"[MeanShift] iteration: {iter}/{self.max_iter}", end="")
                print(f" with delta-centers-norm {diff}")

            if diff < self.tol:
                if self.verbose:
                    print(f"[MeanShift] Early-convergence occurred at", end="")
                    print(f" iteration {iter}/{self.max_iter}")
                break
            centers = new_centers

        self.centers = centers
        self._fitted = True
        return self

    @property
    def labels(self) -> Vector:
        return self.predict(self._X)

    def predict(self, X: Matrix) -> Vector:
        norm_ = np.linalg.norm(X[:, np.newaxis] - self.centers, axis=2)
        cluster_labels = np.argmin(norm_, axis=1)

        unique_labels = np.unique(cluster_labels[cluster_labels != -1])
        mapping = {original: new for new, original in enumerate(unique_labels)}
        labels = [mapping[label] if label != -1 else -1 for label in cluster_labels]

        return Vector(labels)

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)



from typing import List, Literal
import numpy as np

from luma.core.super import Estimator, Evaluator, Supervised
from luma.interface.util import Matrix, Vector, Scalar, Clone
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.metric.classification import Accuracy
from luma.metric.regression import MeanSquaredError
from luma.preprocessing.encoder import LabelEncoder


__all__ = ("VotingClassifier", "VotingRegressor")


class VotingClassifier(Estimator, Estimator.Meta, Supervised):
    """
    A Voting Classifier is an ensemble machine learning model that combines
    predictions from multiple base models. It uses a voting mechanism, either
    'hard' (majority rule) or 'soft' (weighted probabilities), to determine
    the final output. This approach can increase prediction accuracy and
    robustness by leveraging the strengths of diverse models. It's
    particularly useful when individual models have different error patterns.
    The Voting Classifier is widely used in classification tasks to improve
    performance over single models.

    Parameters
    ----------
    `estimators` : List of estimators to vote
    `voting` : Voting criterion
    (`label` to vote the most frequent and `prob` to vote the most probable)
    `weights` : Weights for each classifier on voting
    (`None` for uniform weights)

    Examples
    --------
    >>> vot = VotingClassifier(estimators=[AnyClassifier(), ...],
                               voting='label',
                               weights=[0.25, 0.5, ...])
    >>> vot.fit(X, y)
    >>> pred = vot.predict(X)
    >>> clf = vot[i] # Get i-th classifier from `vot`

    """

    def __init__(
        self,
        estimators: List[Estimator] = None,
        voting: Literal["label", "prob"] = "label",
        weights: Vector[Scalar] = None,
        verbose: bool = False,
    ) -> None:
        self.estimators = estimators
        self.voting = voting
        self.weights = weights
        self.verbose = verbose
        self._fitted = False

    def fit(self, X: Matrix, y: Vector) -> "VotingClassifier":
        if self.voting not in ("label", "prob"):
            raise UnsupportedParameterError(self.voting)

        if self.weights and len(self.weights) != len(self.estimators):
            raise ValueError(f"Size mismatch between 'weights' and 'estimators'!")

        self.le = LabelEncoder()
        self.le.fit(y)

        self.classes = self.le.classes
        self.estimators_ = []

        for est in self.estimators:
            fitted_est = Clone(est).get
            if hasattr(fitted_est, "verbose"):
                fitted_est.verbose = self.verbose

            fitted_est.fit(X, self.le.transform(y))
            self.estimators_.append(fitted_est)

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Vector:
        if not self._fitted:
            raise NotFittedError(self)

        if self.voting == "label":
            preds = np.asarray([est.predict(X) for est in self.estimators_]).T
            majority_vote = np.apply_along_axis(
                lambda x: np.argmax(np.bincount(x, weights=self.weights)),
                axis=1,
                arr=preds,
            )
        else:
            majority_vote = np.argmax(self.predict_proba(X), axis=1)

        majority_vote = self.le.inverse_transform(majority_vote)
        return majority_vote

    def predict_proba(self, X: Matrix) -> Matrix:
        probas = []
        for est in self.estimators_:
            if not hasattr(est, "predict_proba"):
                raise ValueError(f"'{type(est)}' does not support 'predict_proba'")
            probas.append(est.predict_proba(X))

        return np.average(probas, axis=0, weights=self.weights)

    def score(self, X: Matrix, y: Vector, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]


class VotingRegressor(Estimator, Estimator.Meta, Supervised):
    """
    A Voting Regressor is an ensemble machine learning model that combines
    predictions from multiple base regression models. It uses a voting mechanism,
    where the final output is the average of the predictions from all models.
    This approach can increase prediction accuracy and robustness by leveraging
    the strengths of diverse models. It's particularly useful when individual
    models have different error patterns. The Voting Regressor is widely used in
    regression tasks to improve performance over single models.

    Parameters
    ----------
    `estimators` : List of regressors to vote
    `weights` : Weights for each regressor on voting
    (`None` for uniform weights)

    Examples
    --------
    >>> vot = VotingRegressor(estimators=[AnyRegressor(), ...],
                              weights=[0.25, 0.5, ...])
    >>> vot.fit(X, y)
    >>> pred = vot.predict(X)
    >>> reg = vot[i] # Get i-th regressor from `vot`

    """

    def __init__(
        self,
        estimators: List[Estimator] = None,
        weights: Vector[Scalar] = None,
        verbose: bool = False,
    ) -> None:
        self.estimators = estimators
        self.weights = weights
        self.verbose = verbose
        self._fitted = False

    def fit(self, X: Matrix, y: Vector) -> "VotingRegressor":
        if self.weights and len(self.weights) != len(self.estimators):
            raise ValueError(f"Size mismatch between 'weights' and 'estimators'!")

        self.estimators_ = []
        for est in self.estimators:
            fitted_est = Clone(est).get
            if hasattr(fitted_est, "verbose"):
                fitted_est.verbose = self.verbose

            fitted_est.fit(X, y)
            self.estimators_.append(fitted_est)

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Vector:
        if not self._fitted:
            raise NotFittedError(self)

        predictions = np.asarray([est.predict(X) for est in self.estimators_])
        avg_predictions = np.average(predictions, axis=0, weights=self.weights)
        return avg_predictions

    def score(
        self, X: Matrix, y: Vector, metric: Evaluator = MeanSquaredError
    ) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]



from typing import Literal
import pickle
import os

from luma.core.super import Estimator, Transformer
from luma.interface.exception import ModelExtensionError, UnsupportedParameterError

MODEL_EXTENSION = ".luma"


__all__ = "ModelPorter"


class ModelPorter:
    """
    A utility class for exporting and importing machine learning models,
    specifically designed to handle `Estimator` and `Transformer` instances
    within the `luma` framework. It supports saving models to disk with
    automatic naming and enforcing model extension, as well as loading models
    ensuring the correct file extension is used.

    Methods
    -------
    To save(export) models:
    ```py
    def save(self, model, path, filename, replace) -> str
    ```
    It saves a given model to the specified path with an optional filename.
    If the model is not yet fitted, a warning will be printed. The method
    enforces the use of the correct model file extension and handles file
    existence checks based on the `replace` parameter.

    To load(import) models:
    ```py
    def load(self, filepath) -> Estimator | Transformer
    ```
    Loads a model from the given file path, ensuring that the file has the
    correct model extension(`.luma`) before proceeding.

    Raises
    ------
    - `ModelExtensionError` : If the file does not have `.luma` file extension.

    Examples
    --------
    ```py
    model = AnyModel()
    port = ModelPorter()

    model_path = port.save(model=model, path='...', filename='auto') # Export
    new_model = port.load(filepath=model_path) # Import
    ```
    """

    def __init__(self) -> None:
        self._import_filepath = None
        self._export_filepath = None
        self._model_in: Estimator | Transformer = None
        self._model_out: Estimator | Transformer = None

    def save(
        self,
        model: Estimator | Transformer,
        path: str,
        filename: str | Literal["auto"] = "auto",
        replace: bool = False,
    ) -> str:
        if not isinstance(model, (Estimator, Transformer)):
            raise UnsupportedParameterError(model)
        self._model_out = model

        if not self._model_out._fitted:
            print(
                f"[ModelPorter] {type(self._model_out).__name__}",
                "is not fitted! Saving unfitted model.",
            )

        path += "/" if path[-1] != "/" else ""
        self._export_filepath = path

        if filename == "auto":
            self._export_filepath += type(self._model_out).__name__
        else:
            self._export_filepath += filename

        if MODEL_EXTENSION not in filename:
            self._export_filepath += MODEL_EXTENSION

        if os.path.exists(self._export_filepath) and not replace:
            raise FileExistsError(
                f"'{self._export_filepath}' already exists!"
                + " Change 'replace' to 'True' to override."
            )

        with open(self._export_filepath, "wb") as file_out:
            pickle.dump(model, file_out)

        return self._export_filepath

    def load(self, filepath: str) -> Estimator | Transformer:
        self._import_filepath = filepath
        if self._import_filepath[-5:] != MODEL_EXTENSION:
            raise ModelExtensionError(self._import_filepath)

        with open(self._import_filepath, "rb") as file_in:
            self._model_in = pickle.load(file_in)

        return self._model_in



from luma.core.main import Luma
from luma.core.base import ModelBase, ParadigmBase, MetricBase, VisualBase
from luma.core.super import Estimator, Transformer, Optimizer, Evaluator, Visualizer
from luma.core.super import Supervised, Unsupervised, Distance

from luma.interface.exception import NotFittedError, NotConvergedError
from luma.interface.exception import (
    UnsupportedParameterError,
    ModelExtensionError,
    InvalidRangeError,
)
from luma.interface.util import (
    Matrix,
    Vector,
    Scalar,
    DecisionTreeNode,
    NearestNeighbors,
)
from luma.interface.util import SilhouetteUtil, DBUtil, KernelUtil, ActivationUtil
from luma.interface.util import Clone, ParamRange

from luma.classifier.discriminant import (
    LDAClassifier,
    QDAClassifier,
    RDAClassifier,
    KDAClassifier,
)
from luma.classifier.logistic import LogisticRegressor, SoftmaxRegressor
from luma.classifier.naive_bayes import GaussianNaiveBayes, BernoulliNaiveBayes
from luma.classifier.svm import SVC, KernelSVC
from luma.classifier.tree import DecisionTreeClassifier
from luma.classifier.neighbors import (
    KNNClassifier,
    AdaptiveKNNClassifier,
    WeightedKNNClassifier,
)

from luma.clustering.kmeans import (
    KMeansClustering,
    KMeansClusteringPlus,
    KMediansClustering,
)
from luma.clustering.kmeans import KMedoidsClustering, MiniBatchKMeansClustering
from luma.clustering.kmeans import FuzzyCMeansClustering
from luma.clustering.hierarchy import AgglomerativeClustering, DivisiveClustering
from luma.clustering.spectral import SpectralClustering, NormalizedSpectralClustering
from luma.clustering.spectral import (
    HierarchicalSpectralClustering,
    AdaptiveSpectralClustering,
)
from luma.clustering.density import DBSCAN, OPTICS, DENCLUE, MeanShiftClustering
from luma.clustering.affinity import AffinityPropagation, AdaptiveAffinityPropagation
from luma.clustering.affinity import KernelAffinityPropagation
from luma.clustering.mixture import GaussianMixture, MultinomialMixture

from luma.ensemble.forest import RandomForestClassifier, RandomForestRegressor
from luma.ensemble.vote import VotingClassifier, VotingRegressor
from luma.ensemble.bagging import BaggingClassifier, BaggingRegressor
from luma.ensemble.boost import AdaBoostClassifier, AdaBoostRegressor
from luma.ensemble.boost import GradientBoostingClassifier, GradientBoostingRegressor
from luma.ensemble.stack import StackingClassifier, StackingRegressor

from luma.neural.activation import ReLU, LeakyReLU, ELU, Tanh, Sigmoid, Softmax
from luma.neural.optimizer import SGDOptimizer, MomentumOptimizer, RMSPropOptimizer
from luma.neural.single import PerceptronClassifier, PerceptronRegressor
from luma.neural.multi_layer import MLPClassifier

from luma.metric.classification import Accuracy, Precision, Recall, F1Score, Specificity
from luma.metric.regression import (
    MeanAbsoluteError,
    MeanSquaredError,
    RootMeanSquaredError,
    MeanAbsolutePercentageError,
    RSquaredScore,
    AdjustedRSquaredScore,
)
from luma.metric.clustering import SilhouetteCoefficient, DaviesBouldin, Inertia
from luma.metric.distance import Euclidean, Manhattan, Chebyshev, Minkowski
from luma.metric.distance import CosineSimilarity, Correlation, Mahalanobis

from luma.model_selection.split import TrainTestSplit
from luma.model_selection.search import GridSearchCV, RandomizedSearchCV
from luma.model_selection.cv import CrossValidator
from luma.model_selection.fold import KFold, StratifiedKFold

from luma.preprocessing.scaler import StandardScaler, MinMaxScaler
from luma.preprocessing.encoder import OneHotEncoder, LabelEncoder, OrdinalEncoder
from luma.preprocessing.encoder import LabelBinarizer
from luma.preprocessing.imputer import SimpleImputer, KNNImputer, HotDeckImputer
from luma.preprocessing.outlier import LocalOutlierFactor

from luma.reduction.linear import PCA, KernelPCA, LDA, KDA, CCA
from luma.reduction.linear import TruncatedSVD, FactorAnalysis
from luma.reduction.manifold import TSNE, SammonMapping, LaplacianEigenmap
from luma.reduction.manifold import MDS, MetricMDS, LandmarkMDS
from luma.reduction.manifold import LLE, ModifiedLLE, HessianLLE, LTSA
from luma.reduction.manifold import Isomap, ConformalIsomap
from luma.reduction.selection import SBS, SFS, RFE

from luma.regressor.linear import LinearRegressor, LassoRegressor, RidgeRegressor
from luma.regressor.linear import (
    ElasticNetRegressor,
    KernelRidgeRegressor,
    BayesianRidgeRegressor,
)
from luma.regressor.poly import PolynomialRegressor
from luma.regressor.general import PoissonRegressor
from luma.regressor.general import NegativeBinomialRegressor
from luma.regressor.general import GammaRegressor
from luma.regressor.general import BetaRegressor
from luma.regressor.general import InverseGaussianRegressor
from luma.regressor.svm import SVR, KernelSVR
from luma.regressor.tree import DecisionTreeRegressor
from luma.regressor.neighbors import (
    KNNRegressor,
    AdaptiveKNNRegressor,
    WeightedKNNRegressor,
)
from luma.regressor.robust import RANSAC, MLESAC

from luma.pipe.pipeline import Pipeline

from luma.visual.eda import (
    CorrelationBar,
    CorrelationHeatMap,
    JointPlot,
    MissingProportion,
)
from luma.visual.evaluation import (
    DecisionRegion,
    ClusterPlot,
    ROCCurve,
    PrecisionRecallCurve,
)
from luma.visual.evaluation import ConfusionMatrix, ResidualPlot, LearningCurve
from luma.visual.evaluation import ValidationCurve, ValidationHeatmap, InertiaPlot

from luma.migrate.port import ModelPorter


if __name__ == "__main__":

    # ------------------- [ luma.core ] ------------------------
    Luma

    ModelBase, ParadigmBase, MetricBase, VisualBase

    Estimator, Transformer, Optimizer, Evaluator, Visualizer,
    Supervised, Unsupervised, Distance

    # ----------------- [ luma.interface ] ---------------------
    NotFittedError, NotConvergedError,
    UnsupportedParameterError, ModelExtensionError,
    InvalidRangeError

    Matrix, Vector, Scalar, DecisionTreeNode, NearestNeighbors,
    SilhouetteUtil, DBUtil, KernelUtil, ActivationUtil,
    Clone, ParamRange

    # ----------------- [ luma.classifier ] --------------------
    LDAClassifier, QDAClassifier, RDAClassifier, KDAClassifier

    LogisticRegressor, SoftmaxRegressor

    GaussianNaiveBayes, BernoulliNaiveBayes

    SVC, KernelSVC

    DecisionTreeClassifier

    KNNClassifier, AdaptiveKNNClassifier, WeightedKNNClassifier

    # ----------------- [ luma.clustering ] --------------------
    KMeansClustering, KMeansClusteringPlus, KMediansClustering,
    KMedoidsClustering, MiniBatchKMeansClustering,
    FuzzyCMeansClustering

    AgglomerativeClustering, DivisiveClustering

    SpectralClustering, NormalizedSpectralClustering,
    HierarchicalSpectralClustering, AdaptiveSpectralClustering

    DBSCAN, OPTICS, DENCLUE, MeanShiftClustering

    AffinityPropagation, AdaptiveAffinityPropagation,
    KernelAffinityPropagation

    GaussianMixture, MultinomialMixture

    # ------------------ [ luma.ensemble ] ---------------------
    RandomForestClassifier, RandomForestRegressor

    VotingClassifier, VotingRegressor

    BaggingClassifier, BaggingRegressor

    AdaBoostClassifier, AdaBoostRegressor,
    GradientBoostingClassifier, GradientBoostingRegressor

    StackingClassifier, StackingRegressor

    # ------------------- [ luma.neural ] ----------------------
    PerceptronClassifier, PerceptronRegressor

    MLPClassifier

    ReLU, LeakyReLU, ELU, Tanh, Sigmoid, Softmax

    SGDOptimizer, MomentumOptimizer, RMSPropOptimizer

    # ------------------- [ luma.metric ] ----------------------
    Accuracy, Precision, Recall, F1Score, Specificity

    MeanAbsoluteError, MeanSquaredError, RootMeanSquaredError,
    MeanAbsolutePercentageError, RSquaredScore,
    AdjustedRSquaredScore

    SilhouetteCoefficient, DaviesBouldin, Inertia

    Euclidean, Manhattan, Chebyshev, Minkowski,
    CosineSimilarity, Correlation, Mahalanobis

    # --------------- [ luma.module_selection ] ----------------
    TrainTestSplit

    GridSearchCV, RandomizedSearchCV

    CrossValidator

    KFold, StratifiedKFold

    # ---------------- [ luma.preprocessing ] ------------------
    StandardScaler, MinMaxScaler

    OneHotEncoder, LabelEncoder, OrdinalEncoder, LabelBinarizer

    SimpleImputer, KNNImputer, HotDeckImputer

    LocalOutlierFactor

    # ----------------- [ luma.reduction ] ---------------------
    PCA, KernelPCA, LDA, KDA, CCA, TruncatedSVD, FactorAnalysis

    TSNE, SammonMapping, LaplacianEigenmap,
    MDS, MetricMDS, LandmarkMDS,
    LLE, ModifiedLLE, HessianLLE, LTSA,
    Isomap, ConformalIsomap

    SBS, SFS, RFE

    # ----------------- [ luma.regressor ] ---------------------
    LinearRegressor, RidgeRegressor, LassoRegressor,
    ElasticNetRegressor, KernelRidgeRegressor,
    BayesianRidgeRegressor

    PolynomialRegressor

    PoissonRegressor, NegativeBinomialRegressor, GammaRegressor,
    BetaRegressor, InverseGaussianRegressor

    SVR, KernelSVR

    DecisionTreeRegressor

    KNNRegressor, AdaptiveKNNRegressor, WeightedKNNRegressor

    RANSAC, MLESAC

    # -------------------- [ luma.pipe ] -----------------------
    Pipeline

    # ------------------- [ luma.visual ] ----------------------
    CorrelationBar, CorrelationHeatMap, JointPlot,
    MissingProportion

    DecisionRegion, ClusterPlot, ROCCurve, PrecisionRecallCurve,
    ConfusionMatrix, ResidualPlot, LearningCurve,
    ValidationCurve, ValidationHeatmap, InertiaPlot

    # ------------------ [ luma.migrate ] ----------------------
    ModelPorter



from typing import Literal
import numpy as np

from luma.core.super import Transformer
from luma.interface.util import Matrix, Vector
from luma.interface.exception import NotFittedError, UnsupportedParameterError


__all__ = ("OneHotEncoder", "LabelEncoder", "OrdinalEncoder", "LabelBinarizer")


class OneHotEncoder(Transformer, Transformer.Feature):
    def __init__(self, features: list = None):
        self.categories = None
        self.features = features
        self._fitted = False

    def fit(self, X: Matrix) -> "OneHotEncoder":
        if self.features is None:
            self.features = range(X.shape[1])

        self.categories = [np.unique(X[:, col]) for col in self.features]
        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix[int]:
        if not self._fitted:
            raise NotFittedError(self)

        X_out = []
        for i, col in enumerate(X.T):
            if i in self.features:
                categories = self.categories[self.features.index(i)]
                label_to_index = {label: idx for idx, label in enumerate(categories)}
                matrix = np.zeros((len(X), len(categories)))

                for j, item in enumerate(col):
                    if item in label_to_index:
                        matrix[j, label_to_index[item]] = 1
                    elif item is np.nan:
                        continue
                    else:
                        raise ValueError(f"Unknown label {item} found in column {i}")
                X_out.append(matrix)

            else:
                X_out.append(X[:, i].reshape(-1, 1))

        return np.hstack(X_out).astype(int)

    def fit_transform(self, X: Matrix) -> Matrix[int]:
        self.fit(X)
        return self.transform(X)


class LabelEncoder(Transformer, Transformer.Target):
    def __init__(self):
        self.classes = None
        self._fitted = False

    def fit(self, y: Matrix) -> "LabelEncoder":
        self.classes = np.unique(y)
        self._fitted = True
        return self

    def transform(self, y: Matrix) -> Matrix[int]:
        if not self._fitted:
            raise NotFittedError(self)
        class_to_index = {k: v for v, k in enumerate(self.classes)}

        X_transformed = Matrix([class_to_index.get(y_, -1) for y_ in y])
        if -1 in X_transformed:
            raise ValueError("Unknown label found in input data.")

        return X_transformed

    def inverse_transform(self, y: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        index_to_class = {v: k for k, v in enumerate(self.classes)}

        X_inversed = Matrix([index_to_class.get(y_, None) for y_ in y])
        if None in X_inversed:
            raise ValueError("Unknown index found in input data.")

        return X_inversed

    def fit_transform(self, y: Matrix) -> Matrix[int]:
        self.fit(y)
        return self.transform(y)


class OrdinalEncoder(Transformer, Transformer.Feature):
    def __init__(self, strategy: Literal["occur", "alpha"] = "occur"):
        self.categories = None
        self.strategy = strategy
        self._fitted = False

    def fit(self, X: Matrix) -> "OrdinalEncoder":
        if self.strategy == "occur":
            self.categories = [np.unique(col, return_index=True)[0] for col in X.T]
        elif self.strategy == "alpha":
            self.categories = [np.sort(np.unique(col)) for col in X.T]
        else:
            raise UnsupportedParameterError(self.strategy)

        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix[int]:
        if not self._fitted:
            raise NotFittedError(self)
        X_out = np.zeros(X.shape, dtype=int)

        for i, categories in enumerate(self.categories):
            category_to_index = {
                category: index for index, category in enumerate(categories)
            }
            for j, item in enumerate(X[:, i]):
                if item in category_to_index:
                    X_out[j, i] = category_to_index[item]
                else:
                    raise ValueError(f"Unknown label {item} found in column {i}")

        return X_out

    def fit_transform(self, X: Matrix) -> Matrix[int]:
        self.fit(X)
        return self.transform(X)


class LabelBinarizer(Transformer, Transformer.Target):
    def __init__(self, negative_target: bool = False) -> None:
        self.negative_target = negative_target
        self.classes_ = None
        self._fitted = False

    def fit(self, y: Vector) -> "LabelBinarizer":
        self.classes_ = np.unique(y)
        self._fitted = True
        return self

    def transform(self, y: Vector) -> Vector:
        if not self._fitted:
            raise NotFittedError(self)
        if self.negative_target:
            binarized = np.full((len(y), len(self.classes_)), -1)
        else:
            binarized = np.full((len(y), len(self.classes_)), 0)

        for i, label in enumerate(y):
            class_index = np.where(self.classes_ == label)[0][0]
            binarized[i, class_index] = 1

        return binarized

    def fit_transform(self, y: Vector) -> Vector:
        self.fit(y)
        return self.transform(y)

    def inverse_transform(self, y: Vector) -> Vector:
        return self.classes_[np.argmax(y, axis=1)]



from typing import Literal
from scipy.stats import mode
import numpy as np

from luma.interface.util import Matrix, Vector
from luma.classifier.tree import DecisionTreeClassifier
from luma.regressor.tree import DecisionTreeRegressor
from luma.core.super import Estimator, Evaluator, Supervised
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.metric.classification import Accuracy
from luma.metric.regression import MeanSquaredError


__all__ = ("RandomForestClassifier", "RandomForestRegressor")


class RandomForestClassifier(Estimator, Estimator.Meta, Supervised):
    """
    A Random Forest Classifier is an ensemble learning method in machine learning,
    widely used for both classification and regression tasks. It operates by
    constructing a multitude of decision trees during the training phase and
    outputs the mode of the classes for classification.

    Parameters
    ----------
    `n_trees` : Number of trees in the forest
    `max_depth` :  Maximum depth of each trees
    `criterion` : Function used to measure the quality of a split
    `min_samples_split` : Minimum samples required to split a node
    `min_samples_leaf` : Minimum samples required to be at a leaf node
    `max_features` : Number of features to consider
    `min_impurity_decrease` : Minimum decrement of impurity for a split
    `max_leaf_nodes` : Maximum amount of leaf nodes
    `bootstrap` : Whether to bootstrap the samples of dataset
    `bootstrap_feature` : Whether to bootstrap the features of each data
    `n_features` : Number of features to be sampled when bootstrapping features

    """

    def __init__(
        self,
        n_trees: int = 10,
        max_depth: int = 100,
        criterion: Literal["gini", "entropy"] = "gini",
        min_samples_split: int = 2,
        min_samples_leaf: int = 1,
        max_features: int = None,
        min_impurity_decrease: float = 0.0,
        max_leaf_nodes: int = None,
        bootstrap: bool = True,
        bootstrap_feature: bool = False,
        n_features: int | Literal["auto"] = "auto",
        verbose: bool = False,
    ) -> None:
        self.n_trees = n_trees
        self.max_depth = max_depth
        self.criterion = criterion
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.min_impurity_decrease = min_impurity_decrease
        self.max_leaf_nodes = max_leaf_nodes
        self.n_features = n_features
        self.bootstrap = bootstrap
        self.bootstrap_feature = bootstrap_feature
        self.verbose = verbose
        self.trees = None
        self._fitted = False

        self.set_param_ranges(
            {
                "n_trees": ("0<,+inf", int),
                "max_depth": ("0<,+inf", int),
                "min_samples_split": ("0<,+inf", int),
                "max_samples_leaf": ("0<,+inf", int),
                "min_impurity_decrease": ("0,+inf", None),
            }
        )
        self.check_param_ranges()

    def fit(
        self, X: Matrix, y: Matrix, sample_weights: Vector = None
    ) -> "RandomForestClassifier":
        m, n = X.shape
        if self.n_features == "auto":
            self.n_features = int(n**0.5)
        else:
            if isinstance(self.n_features, str):
                raise UnsupportedParameterError(self.n_features)

        _tree_params = {
            "max_depth": self.max_depth,
            "criterion": self.criterion,
            "min_samples_split": self.min_samples_split,
            "min_samples_leaf": self.min_samples_leaf,
            "max_features": self.max_features,
            "min_impurity_decrease": self.min_impurity_decrease,
            "max_leaf_nodes": self.max_leaf_nodes,
        }

        self.trees = [DecisionTreeClassifier() for _ in range(self.n_trees)]
        for i, tree in enumerate(self.trees, start=1):
            X_sample, y_sample = X, y
            if self.bootstrap:
                bootstrap_data = np.random.choice(m, m, replace=True)
                X_sample = X_sample[bootstrap_data]
                y_sample = y_sample[bootstrap_data]
            if self.bootstrap_feature:
                bootstrap_feature = np.random.choice(n, self.n_features, replace=True)
                X_sample = X_sample[:, bootstrap_feature]

            tree.set_params(**_tree_params)
            tree.fit(X_sample, y_sample, sample_weights=sample_weights)

            if self.verbose:
                print(f"[RandomForest] Finished fitting tree {i}/{self.n_trees}")

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        preds = Matrix([tree.predict(X) for tree in self.trees])
        majority, _ = mode(preds, axis=0)

        return majority.flatten()

    def score(self, X: Matrix, y: Matrix, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def __gettiem__(self, index: int) -> DecisionTreeClassifier:
        return self.trees[index]


class RandomForestRegressor(Estimator, Estimator.Meta, Supervised):
    """
    A Random Forest Regressor is an ensemble learning algorithm for regression
    tasks. It builds multiple decision trees during training, each on a random
    subset of features and bootstrapped data. The final prediction is the average
    of individual tree predictions, providing a robust and accurate result.

    Parameters
    ----------
    `n_trees` : Number of trees in the forest
    `max_depth` :  Maximum depth of each trees
    `min_samples_split` : Minimum samples required to split a node
    `min_samples_leaf` : Minimum samples required to be at a leaf node
    `max_features` : Number of features to consider
    `min_variance_decrease` : Minimum decrement of variance for a split
    `max_leaf_nodes` : Maximum amount of leaf nodes
    `bootstrap` : Whether to bootstrap the samples of dataset
    `bootstrap_feature` : Whether to bootstrap the features of each data
    `n_features` : Number of features to be sampled when bootstrapping features

    """

    def __init__(
        self,
        n_trees: int = 10,
        max_depth: int = 10,
        min_samples_split: int = 2,
        min_samples_leaf: int = 1,
        max_features: int = None,
        min_variance_decrease: float = 0.0,
        max_leaf_nodes: int = None,
        bootstrap: bool = True,
        bootstrap_feature: bool = False,
        n_features: int | Literal["auto"] = "auto",
        sample_weights: Vector = None,
        verbose: bool = False,
    ) -> None:
        self.n_trees = n_trees
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.min_variance_decrease = min_variance_decrease
        self.max_leaf_nodes = max_leaf_nodes
        self.n_features = n_features
        self.bootstrap = bootstrap
        self.bootstrap_feature = bootstrap_feature
        self.sample_weights = sample_weights
        self.verbose = verbose
        self.trees = None
        self._fitted = False

        self.set_param_ranges(
            {
                "n_trees": ("0<,+inf", int),
                "max_depth": ("0<,+inf", int),
                "min_samples_split": ("0<,+inf", int),
                "max_samples_leaf": ("0<,+inf", int),
                "min_variance_decrease": ("0,+inf", None),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Matrix) -> "RandomForestRegressor":
        m, n = X.shape
        if self.n_features == "auto":
            self.n_features = int(n**0.5)
        else:
            if isinstance(self.n_features, str):
                raise UnsupportedParameterError(self.n_features)

        _tree_params = {
            "max_depth": self.max_depth,
            "min_samples_split": self.min_samples_split,
            "min_samples_leaf": self.min_samples_leaf,
            "max_features": self.max_features,
            "min_variance_decrease": self.min_variance_decrease,
            "max_leaf_nodes": self.max_leaf_nodes,
        }

        self.trees = [DecisionTreeRegressor() for _ in range(self.n_trees)]
        for i, tree in enumerate(self.trees, start=1):
            X_sample, y_sample = X, y
            if self.bootstrap:
                bootstrap_data = np.random.choice(m, m, replace=True)
                X_sample = X_sample[bootstrap_data]
                y_sample = y_sample[bootstrap_data]
            if self.bootstrap_feature:
                bootstrap_feature = np.random.choice(n, self.n_features, replace=True)
                X_sample = X_sample[:, bootstrap_feature]

            tree.set_params(**_tree_params)
            tree.fit(X_sample, y_sample, sample_weights=self.sample_weights)

            if self.verbose:
                print(f"[RandomForest] Finished fitting tree {i}/{self.n_trees}")

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        preds = Matrix([tree.predict(X) for tree in self.trees])
        average_predictions = np.mean(preds, axis=0)

        return average_predictions

    def score(
        self, X: Matrix, y: Matrix, metric: Evaluator = MeanSquaredError
    ) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def __gettiem__(self, index: int) -> DecisionTreeClassifier:
        return self.trees[index]



from itertools import product
from typing import Any, Dict, List, Literal, Optional, Tuple
from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt
import numpy as np

from luma.interface.util import Matrix, Vector
from luma.core.super import Evaluator, Visualizer, Estimator
from luma.model_selection.split import TrainTestSplit
from luma.preprocessing.encoder import LabelBinarizer
from luma.metric.classification import Accuracy, Precision, Recall, Specificity
from luma.model_selection.cv import CrossValidator
from luma.model_selection.fold import FoldType, KFold


__all__ = (
    "DecisionRegion",
    "ClusterPlot",
    "ROCCurve",
    "PrecisionRecallCurve",
    "ConfusionMatrix",
    "ResidualPlot",
    "LearningCurve",
    "ValidationCurve",
    "ValidationHeatmap",
    "InertiaPlot",
)


class DecisionRegion(Visualizer):
    def __init__(
        self,
        estimator: Estimator,
        X: Matrix,
        y: Optional[Matrix] = None,
        title: str | Literal["auto"] = "auto",
        xlabel: str = r"$x_1$",
        ylabel: str = r"$x_2$",
        cmap: ListedColormap = "rainbow",
        alpha: float = 0.4,
    ) -> None:
        self.estimator = estimator
        self.X = X
        self.y = y
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.cmap = cmap
        self.alpha = alpha

        if self.title == "auto":
            self.title = type(self.estimator).__name__

        if self.y is None and hasattr(self.estimator, "labels"):
            self.y = self.estimator.labels

    def plot(
        self,
        ax: Optional[plt.Axes] = None,
        size: float = 250,
        scale: float = 10.0,
        show: bool = False,
    ) -> plt.Axes:
        x1_min, x1_max = self.X[:, 0].min(), self.X[:, 0].max()
        x2_min, x2_max = self.X[:, 1].min(), self.X[:, 1].max()
        delta_1, delta_2 = (x1_max - x1_min) / size, (x2_max - x2_min) / size

        x1_min, x1_max = x1_min - delta_1 * scale, x1_max + delta_1 * scale
        x2_min, x2_max = x2_min - delta_2 * scale, x2_max + delta_2 * scale
        xx1, xx2 = np.meshgrid(
            np.arange(x1_min, x1_max, delta_1), np.arange(x2_min, x2_max, delta_2)
        )

        Z = self.estimator.predict(Matrix([xx1.ravel(), xx2.ravel()]).T)
        Z = Z.reshape(xx1.shape)

        if ax is None:
            _, ax = plt.subplots()
            show = True

        ax.contourf(
            xx1, xx2, Z, alpha=self.alpha, cmap=self.cmap, levels=len(np.unique(self.y))
        )

        ax.set_xlim(xx1.min(), xx1.max())
        ax.set_ylim(xx2.min(), xx2.max())

        ax.scatter(
            self.X[:, 0],
            self.X[:, 1],
            c=self.y,
            cmap=self.cmap,
            alpha=0.8,
            edgecolors="black",
        )

        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)
        ax.set_title(self.title)
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax


class ClusterPlot(Visualizer):
    def __init__(
        self,
        estimator: Estimator,
        X: Matrix,
        title: str | Literal["auto"] = "auto",
        xlabel: str = r"$x_1$",
        ylabel: str = r"$x_2$",
        cmap: ListedColormap = "rainbow",
        alpha: float = 0.8,
    ) -> None:
        self.estimator = estimator
        self.X = X
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.cmap = cmap
        self.alpha = alpha
        self.labels = self.estimator.labels

        if self.title == "auto":
            self.title = type(self.estimator).__name__

    def plot(self, ax: Optional[plt.Axes] = None, show: bool = False) -> plt.Axes:
        if ax is None:
            _, ax = plt.subplots()
            show = True

        ax.scatter(
            self.X[self.labels == -1, 0],
            self.X[self.labels == -1, 1],
            marker="x",
            c="black",
            label="Noise",
        )

        ax.scatter(
            self.X[self.labels != -1, 0],
            self.X[self.labels != -1, 1],
            marker="o",
            c=self.labels[self.labels != -1],
            cmap=self.cmap,
            alpha=self.alpha,
            edgecolors="black",
        )

        ax.set_title(self.title)
        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)

        if len(self.X[self.labels == -1]):
            ax.legend()
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax


class ROCCurve(Visualizer):
    def __init__(self, y_true: Vector, y_scores: Matrix) -> None:
        self.y_true = y_true
        self.y_scores = y_scores
        self.n_classes = y_scores.shape[1]

    def plot(self, ax: Optional[plt.Axes] = None, show: bool = False) -> plt.Axes:
        if ax is None:
            _, ax = plt.subplots()
            show = True

        y_binary = LabelBinarizer().fit_transform(self.y_true)

        fprs, tprs = [], []
        for cl in range(self.n_classes):
            fpr, tpr = self._fpr_tpr(y_binary[:, cl], self.y_scores[:, cl])
            fprs.append(fpr)
            tprs.append(tpr)

            auc = self._auc(fpr, tpr)
            ax.plot(fpr, tpr, linewidth=2, label=f"ROC Curve {cl} (area = {auc:.2f})")

        mean_fpr, mean_tpr = np.mean(fprs, axis=0), np.mean(tprs, axis=0)
        mean_auc = self._auc(mean_fpr, mean_tpr)

        ax.plot(
            mean_fpr,
            mean_tpr,
            color="dimgray",
            linewidth=2,
            linestyle="--",
            label=f"Mean ROC (area = {mean_auc:.2f})",
        )

        ax.plot(
            [0, 1], [0, 1], color="darkgray", linestyle=":", label="Random Guessing"
        )

        ax.set_xlim([-0.05, 1.05])
        ax.set_ylim([-0.05, 1.05])
        ax.set_xlabel("False Positive Rate")

        ax.set_ylabel("True Positive Rate")
        ax.set_title("ROC Curve")
        ax.legend(loc="lower right")
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax

    def _fpr_tpr(self, y_true: Vector, y_score: Vector) -> Tuple[Vector, Vector]:
        thresholds = np.sort(y_score)[::-1]
        fpr, tpr = [], []
        for threshold in thresholds:
            y_pred = y_score >= threshold
            tpr.append(Recall.score(y_true, y_pred))
            fpr.append(1 - Specificity.score(y_true, y_pred))

        return Vector(fpr), Vector(tpr)

    def _auc(self, fpr: Vector, tpr: Vector) -> float:
        auc = 0
        for i in range(1, len(fpr)):
            auc += (fpr[i] - fpr[i - 1]) * (tpr[i] + tpr[i - 1]) / 2
        return auc


class PrecisionRecallCurve(Visualizer):
    def __init__(self, y_true: Vector, y_scores: Matrix) -> None:
        self.y_true = y_true
        self.y_scores = y_scores
        self.n_classes = y_scores.shape[1]

    def plot(self, ax: Optional[plt.Axes] = None, show: bool = False) -> plt.Axes:
        if ax is None:
            _, ax = plt.subplots()
            show = True

        y_binary = LabelBinarizer().fit_transform(self.y_true)
        pres, recs = [], []
        for cl in range(self.n_classes):
            pre, rec = self._pre_rec(y_binary[:, cl], self.y_scores[:, cl])
            pres.append(pre)
            recs.append(rec)

            ap = self._average_precision(pre, rec)
            ax.plot(rec, pre, linewidth=2, label=f"PR Curve {cl} (AP = {ap:.2f})")

        mean_pre, mean_rec = np.mean(pres, axis=0), np.mean(recs, axis=0)
        mean_ap = self._average_precision(mean_pre, mean_rec)

        ax.plot(
            mean_rec,
            mean_pre,
            color="dimgray",
            linewidth=2,
            linestyle="--",
            label=f"Mean PR (AP = {mean_ap:.2f})",
        )

        ax.plot(
            [1, 0], [0, 1], color="darkgray", linestyle=":", label="Random Guessing"
        )

        ax.set_xlim([-0.05, 1.05])
        ax.set_ylim([-0.05, 1.05])
        ax.set_xlabel("Recall")
        ax.set_ylabel("Precision")
        ax.set_title("Precision-Recall Curve")
        ax.legend(loc="lower left")
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax

    def _pre_rec(self, y_true: Vector, y_score: Vector) -> Tuple[Vector, Vector]:
        thresholds = np.sort(y_score)[::-1]
        pre, rec = [], []
        for threshold in thresholds:
            y_pred = y_score >= threshold
            pre.append(Precision.score(y_true, y_pred))
            rec.append(Recall.score(y_true, y_pred))

        return Vector(pre), Vector(rec)

    def _average_precision(self, pre: Vector, rec: Vector) -> float:
        ap = 0
        for i in range(1, len(rec)):
            ap += (rec[i] - rec[i - 1]) * pre[i]
        return ap


class ConfusionMatrix(Visualizer):
    def __init__(
        self,
        y_true: Vector,
        y_pred: Vector,
        labels: List[str] | None = None,
        title: str | Literal["auto"] = "auto",
        cmap: ListedColormap = "Blues",
    ) -> None:
        self.cmap = cmap
        self.conf_matrix = self._confusion_matrix(y_true, y_pred)
        self.labels = labels
        self.title = title

        if labels is None:
            self.labels = np.arange(len(np.unique(y_true)))

    def plot(self, ax: Optional[plt.Axes] = None, show: bool = False) -> plt.Axes:
        if ax is None:
            _, ax = plt.subplots()
            show = True

        cax = ax.imshow(self.conf_matrix, interpolation="nearest", cmap=self.cmap)
        if self.title == "auto":
            ax.set_title("Confusion Matrix")
        else:
            ax.set_title(self.title)

        plt.colorbar(cax, ax=ax)

        tick_marks = np.arange(len(self.labels))
        ax.set_xticks(tick_marks)
        ax.set_xticklabels(self.labels, rotation=45)
        ax.set_yticks(tick_marks)
        ax.set_yticklabels(self.labels)

        thresh = self.conf_matrix.max() / 2.0
        for i, j in np.ndindex(self.conf_matrix.shape):
            ax.text(
                j,
                i,
                format(self.conf_matrix[i, j], "d"),
                horizontalalignment="center",
                color="white" if self.conf_matrix[i, j] > thresh else "black",
            )

        ax.set_ylabel("True label")
        ax.set_xlabel("Predicted label")
        ax.set_aspect("equal", adjustable="box")
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax

    def _confusion_matrix(self, y_true: Vector, y_pred: Vector) -> Matrix:
        unique_labels = np.unique(np.concatenate((y_true, y_pred)))
        matrix = np.zeros((len(unique_labels), len(unique_labels)), dtype=int)
        for true, pred in zip(y_true, y_pred):
            matrix[true, pred] += 1

        return matrix


class ResidualPlot(Visualizer):
    def __init__(
        self,
        estimator: Estimator,
        X: Matrix,
        y: Vector,
        alpha: float = 0.8,
        cmap: ListedColormap = "RdYlBu",
    ) -> None:
        self.estimator = estimator
        self.X = X
        self.y = y
        self.alpha = alpha
        self.cmap = cmap

    def plot(self, ax: Optional[plt.Axes] = None, show: bool = False) -> plt.Axes:
        if ax is None:
            _, ax = plt.subplots()
            show = True

        resid = self._calculate_residuals()
        cax = ax.scatter(
            self.estimator.predict(self.X),
            resid,
            c=resid,
            s=20,
            cmap=self.cmap,
            alpha=self.alpha,
        )

        ax.axhline(y=0, c="black", lw=2, label="Perfect Fit")

        ax.axhline(y=resid.mean(), c="gray", ls="--", lw=2, label="Average Residual")

        ax.set_xlabel("Predicted Values")
        ax.set_ylabel("Residuals")
        ax.set_title(f"Residual Plot of {type(self.estimator).__name__}")

        ax.figure.colorbar(cax)
        ax.legend()
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax

    def _calculate_residuals(self) -> Vector:
        if not self.estimator._fitted:
            self.estimator.fit(self.X, self.y)

        predictions = self.estimator.predict(self.X)
        residuals = self.y - predictions
        return residuals


class LearningCurve(Visualizer):
    """
    A learning curve in machine learning is a graph that compares the performance
    of a model on training and validation data over a series of training iterations.
    It helps to diagnose problems like overfitting or underfitting by showing how
    the model's error changes as it learns. A steep learning curve indicates rapid
    learning, while a plateau suggests no further learning. The ideal curve shows
    decreasing error on both training and validation datasets, converging to a
    low level.

    Parameters
    ----------
    `estimator` : An estimator to evaluate
    `X` : Feature data for training
    `y`: Target data for training
    `train_sizes` : List of fractional values for each dataset size
    `test_size` : Proportional value for validation set size
    `cv` : Number of times for cross-validation
    `shuffle` : Whether to shuffle the dataset
    `stratify` : Whether to perform stratified split
    `fold_type` : Fold type (Default `KFold`)
    `metric` : Evaluation metric for scoring
    `random_state` : Seed for random sampling upon splitting samples

    """

    def __init__(
        self,
        estimator: Estimator,
        X: Matrix,
        y: Vector,
        train_sizes: Vector = None,
        test_size: float = 0.3,
        cv: int = 5,
        shuffle: bool = True,
        stratify: bool = False,
        fold_type: FoldType = KFold,
        metric: Evaluator = Accuracy,
        random_state: int = None,
        verbose: bool = False,
    ):
        self.estimator = estimator
        self.X = X
        self.y = y
        self.train_sizes = train_sizes
        self.test_size = test_size
        self.cv = cv
        self.shuffle = shuffle
        self.stratify = stratify
        self.fold_type = fold_type
        self.metric = metric
        self.random_state = random_state
        self.verbose = verbose

        if train_sizes is None:
            self.train_sizes = np.linspace(0.1, 1.0, 5)

    def _evaluate(self) -> None:
        m, _ = self.X.shape
        self.train_scores, self.test_scores = [], []

        for train_size in self.train_sizes:
            n_samples = int(train_size * m)

            X_train, X_val, y_train, y_val = TrainTestSplit(
                self.X,
                self.y,
                test_size=n_samples,
                shuffle=self.shuffle,
                stratify=self.stratify,
                random_state=self.random_state,
            ).get

            if n_samples == len(self.X):
                X_train = X_val = self.X
                y_train = y_val = self.y

            self.estimator.fit(X_train, y_train)
            cv_param = (
                self.estimator,
                self.metric,
                self.cv,
                self.fold_type,
                self.shuffle,
                self.random_state,
                self.verbose,
            )

            if self.verbose:
                print(f"Starting CV for training set with {self.cv} folds")

            cv_train = CrossValidator(*cv_param)
            cv_train._fit(X_train, y_train)
            self.train_scores.append(cv_train.test_scores_)

            if self.verbose:
                print(f"Starting CV for validation set with {self.cv} folds")

            cv_test = CrossValidator(*cv_param)
            cv_test._fit(X_val, y_val)
            self.test_scores.append(cv_test.test_scores_)

            if self.verbose:
                print(
                    f"[LearningCurve] Finished evaluating for", f"{n_samples} samples"
                )

        self.train_scores = Matrix(self.train_scores)
        self.test_scores = Matrix(self.test_scores)

    def plot(self, ax: Optional[plt.Axes] = None, show: bool = False) -> plt.Axes:
        self._evaluate()
        self.train_sizes = (self.train_sizes * self.X.shape[0]).astype(int)
        metric_name = self.metric.__name__

        train_mean = self.train_scores.mean(axis=1)
        train_std = self.train_scores.std(axis=1)
        test_mean = self.test_scores.mean(axis=1)
        test_std = self.test_scores.std(axis=1)

        if ax is None:
            _, ax = plt.subplots()
            show = True

        ax.plot(
            self.train_sizes,
            train_mean,
            "o-",
            color="royalblue",
            label=f"Training {metric_name}",
        )
        ax.plot(
            self.train_sizes,
            test_mean,
            "o--",
            color="seagreen",
            label=f"Validation {metric_name}",
        )

        ax.fill_between(
            self.train_sizes,
            train_mean + train_std,
            train_mean - train_std,
            color="royalblue",
            alpha=0.2,
        )

        ax.fill_between(
            self.train_sizes,
            test_mean + test_std,
            test_mean - test_std,
            color="seagreen",
            alpha=0.2,
        )

        ax.set_title(f"Learning Curve")
        ax.set_xlabel("Number of Training Examples")
        ax.set_ylabel(metric_name)
        ax.set_ylim(0.0, 1.1)
        ax.legend()
        ax.grid()
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax


class ValidationCurve(Visualizer):
    """
    A validation curve in machine learning is a graph that compares the performance
    of a model on training and validation data over a range of values for a specific
    hyperparameter. It helps to diagnose the best hyperparameter setting by showing
    how the model's error changes as the value of the hyperparameter changes.

    Parameters
    ----------
    `estimator` : An estimator to evaluate
    `X` : Feature data for training
    `y`: Target data for training
    `param_name` : Name of the hyperparameter to be varied
    `param_range` : Range of values for the hyperparameter
    `cv` : Number of times for cross-validation
    `shuffle` : Whether to shuffle the dataset
    `fold_type` : Fold type (Default `KFold`)
    `metric` : Evaluation metric for scoring
    `random_state` : Seed for random sampling upon splitting samples

    """

    def __init__(
        self,
        estimator: Estimator,
        X: Matrix,
        y: Vector,
        param_name: str,
        param_range: Vector,
        cv: int = 5,
        shuffle: bool = True,
        fold_type: FoldType = KFold,
        metric: Evaluator = Accuracy,
        random_state: int = None,
        verbose: bool = False,
    ):
        self.estimator = estimator
        self.X = X
        self.y = y
        self.param_name = param_name
        self.param_range = param_range
        self.cv = cv
        self.shuffle = shuffle
        self.fold_type = fold_type
        self.metric = metric
        self.random_state = random_state
        self.verbose = verbose

    def _evaluate(self) -> None:
        self.train_scores, self.test_scores = [], []
        for param_value in self.param_range:
            setattr(self.estimator, self.param_name, param_value)

            cv_model = CrossValidator(
                estimator=self.estimator,
                metric=self.metric,
                cv=self.cv,
                shuffle=self.shuffle,
                fold_type=self.fold_type,
                random_state=self.random_state,
                verbose=self.verbose,
            )

            cv_model._fit(self.X, self.y)
            self.train_scores.append(cv_model.train_scores_)
            self.test_scores.append(cv_model.test_scores_)

            if self.verbose:
                print(
                    f"[ValidationCurve] Finished evaluating for",
                    f"{self.param_name}={param_value}",
                )

        self.train_scores = Matrix(self.train_scores)
        self.test_scores = Matrix(self.test_scores)

    def plot(
        self, ax: Optional[plt.Axes] = None, xscale: str = None, show: bool = False
    ) -> plt.Axes:
        self._evaluate()
        metric_name = self.metric.__name__

        train_mean = self.train_scores.mean(axis=1)
        train_std = self.train_scores.std(axis=1)
        test_mean = self.test_scores.mean(axis=1)
        test_std = self.test_scores.std(axis=1)

        if ax is None:
            _, ax = plt.subplots()
            show = True

        ax.plot(
            self.param_range,
            train_mean,
            "o-",
            color="royalblue",
            label=f"Training {metric_name}",
        )
        ax.plot(
            self.param_range,
            test_mean,
            "o--",
            color="seagreen",
            label=f"Validation {metric_name}",
        )

        ax.fill_between(
            self.param_range,
            train_mean + train_std,
            train_mean - train_std,
            color="royalblue",
            alpha=0.2,
        )

        ax.fill_between(
            self.param_range,
            test_mean + test_std,
            test_mean - test_std,
            color="seagreen",
            alpha=0.2,
        )

        ax.set_title(f"Validation Curve for '{self.param_name}'")
        ax.set_xlabel(self.param_name)
        ax.set_ylabel(metric_name)

        ax.set_xscale(xscale if xscale else "linear")
        ax.set_ylim(0.0, 1.1)
        ax.legend()
        ax.grid()
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax


class ValidationHeatmap(Visualizer):
    """
    A validation heatmap in machine learning is a graphical representation that
    illustrates the performance of a model on a validation dataset for different
    combinations of two hyperparameters. Each cell in the heatmap corresponds to a
    specific combination of the hyperparameters' values, and the color intensity
    represents the performance metric score. This visualization is useful for
    identifying the interaction between two hyperparameters and selecting the best
    combination for model tuning.

    Parameters
    ----------
    `estimator` : An estimator to evaluate
    `X` : Feature data for training
    `y` : Target data for training
    `param_dict` : Dictionary with two hyperparameters and their respective ranges
    `cv` : Number of times for cross-validation
    `shuffle` : Whether to shuffle the dataset
    `fold_type` : Fold type (Default `KFold`)
    `metric` : Evaluation metric for scoring
    `random_state` : Seed for random sampling upon splitting samples

    """

    def __init__(
        self,
        estimator: Estimator,
        X: Matrix,
        y: Vector,
        param_dict: Dict[str, Any],
        cv: int = 5,
        shuffle: bool = True,
        fold_type: FoldType = KFold,
        metric: Evaluator = Accuracy,
        random_state: int = None,
        verbose: bool = False,
    ) -> None:
        self.estimator = estimator
        self.X = X
        self.y = y
        self.param_dict = param_dict
        self.cv = cv
        self.shuffle = shuffle
        self.fold_type = fold_type
        self.metric = metric
        self.random_state = random_state
        self.verbose = verbose

        self.names_ = list(self.param_dict.keys())

        self.sizes_ = (
            len(self.param_dict[self.names_[0]]),
            len(self.param_dict[self.names_[1]]),
        )

        self.values_ = (
            self.param_dict[self.names_[0]],
            self.param_dict[self.names_[1]],
        )

        self.scores = np.zeros(self.sizes_)

    def _evaluate(self) -> None:
        param_combs = list(product(*self.param_dict.values()))
        max_iter = len(param_combs)

        if self.verbose:
            print(
                f"Fitting {self.cv} folds for {max_iter} candidates,",
                f"totalling {self.cv * max_iter} fits.\n",
            )

        for idx, params in enumerate(param_combs):
            param_values = dict(zip(self.names_, params))
            self.estimator.set_params(**param_values)

            cv_model = CrossValidator(
                estimator=self.estimator,
                metric=self.metric,
                cv=self.cv,
                shuffle=self.shuffle,
                fold_type=self.fold_type,
                random_state=self.random_state,
                verbose=False,
            )

            _, score = cv_model.score(self.X, self.y)
            i, j = idx // self.sizes_[1], idx % self.sizes_[1]
            self.scores[i, j] = score

            if self.verbose:
                print(
                    f"[ValidationHeatmap] candidate {idx + 1}/{max_iter}",
                    f"{param_values} - score: {score:.3f}",
                )

    def plot(
        self,
        ax: Optional[plt.Axes] = None,
        cmap: ListedColormap = "RdYlGn",
        log_xticks: bool = True,
        log_yticks: bool = True,
        annotate: bool = True,
        show: bool = False,
    ) -> plt.Axes:
        if ax is None:
            _, ax = plt.subplots()
            show = True

        cax = ax.imshow(self.scores, cmap=cmap)
        ax.figure.colorbar(cax)

        self._evaluate()
        if annotate:
            for i in range(self.sizes_[0]):
                for j in range(self.sizes_[1]):
                    score = self.scores[i, j]
                    ax.text(
                        j,
                        i,
                        f"{score:.2f}",
                        ha="center",
                        va="center",
                        color="white",
                        alpha=0.7,
                    )

        ax.set_xticks(range(self.sizes_[0]), np.round(self.values_[0], 2))
        ax.set_yticks(range(self.sizes_[1]), np.round(self.values_[1], 2))

        if log_xticks:
            ax.set_xticklabels(
                [r"$10^{" + f"{np.log10(n):.1f}" + r"}$" for n in self.values_[0]]
            )
        if log_yticks:
            ax.set_yticklabels(
                [r"$10^{" + f"{np.log10(n):.1f}" + r"}$" for n in self.values_[1]]
            )

        ax.set_xlabel(self.names_[1])
        ax.set_ylabel(self.names_[0])

        ax.set_title("Validation Heatmap")
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax


class InertiaPlot(Visualizer):
    """
    An inertia plot in clustering visualizes the inertia against the number of
    clusters, helping to determine the optimal number of clusters by identifying
    the "elbow" point where the rate of decrease in inertia sharply changes.
    It reflects how well clusters are compact and separated, with lower inertia
    indicating better clustering. The plot aids in applying the elbow method for
    selecting a reasonable number of clusters in algorithms like K-means. The
    goal is to choose the number of clusters where inertia begins to decrease
    more slowly.

    Parameters
    ----------
    `inertia_list` : List of inertia values for various cluster sizes
    `n_clusters_list` : List of the number of clusters

    Examples
    --------
    ```py
    inertia_list = []
    for i in range(2, 10):
        km = KMeansClustering(n_clusters=i)
        km.fit(data)

        inr = Inertia.score(data, km.centroids)
        inertia_list.append(inr)

    inr_plot = InertiaPlot(inertia_list=inertia_list,
                           n_clusters_list=list(range(2, 10)))
    inr_plot.plot()
    >>> plt.Axes
    ```
    """

    def __init__(self, inertia_list: List[float], n_clusters_list: List[int]) -> None:
        self.inertia_list = inertia_list
        self.n_clusters_list = n_clusters_list

    def _derivate_inertia(self) -> Vector:
        deriv = [0.0]
        for i in range(1, len(self.n_clusters_list) - 1):
            df = (self.inertia_list[i - 1] - self.inertia_list[i + 1]) / 2
            deriv.append(df)

        deriv.append(0.0)
        return np.abs(deriv)

    def plot(
        self,
        ax: Optional[plt.Axes] = None,
        marker: str = "o",
        linestyle: str = "-",
        show: bool = False,
    ) -> None:
        if ax is None:
            _, ax = plt.subplots()
            show = True

        ax.plot(
            self.n_clusters_list,
            self.inertia_list,
            marker=marker,
            linestyle=linestyle,
            label="Inertia",
        )

        ax.plot(
            self.n_clusters_list,
            self._derivate_inertia(),
            marker=marker,
            linestyle=linestyle,
            label="Absolute Derivative",
        )

        best_idx = np.argmax(self._derivate_inertia())
        best_n_clusters = self.n_clusters_list[best_idx]
        ax.axvspan(
            best_n_clusters - 0.25,
            best_n_clusters + 0.25,
            color="orange",
            alpha=0.2,
            label="Best Number of Clusters",
        )

        ax.set_title("Inertia Plot")
        ax.set_xlabel("Number of Clusters")
        ax.set_ylabel("Inertia / Deriv. of Inertia")
        ax.legend()
        ax.grid()
        ax.figure.tight_layout()

        if show:
            plt.show()
        return ax
